from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Protocol

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_decimal


@dataclass
class GeometryShape(Protocol):
    color: str

    def to_json_string(self) -> str: ...


@dataclass
class Rectangle(GeometryShape):
    point1: GeometryPoint
    point2: GeometryPoint
    line_width: float
    color: str
    active: bool
    resource_id: str

    def __post_init__(self) -> None:
        _check_warn_shape_info(self.color, self.line_width, self.resource_id)

    def to_json_string(self) -> str:
        json_dict = {
            "status": "active" if self.active else "deleted",
            "type": "rectangle",
            "lineWidth": self.line_width,
            "points": [self.point1.to_dict(), self.point2.to_dict()],
        }
        return json.dumps(json_dict)


@dataclass
class Polygon(GeometryShape):
    points: list[GeometryPoint]
    line_width: float
    color: str
    active: bool
    resource_id: str

    def __post_init__(self) -> None:
        _check_warn_shape_info(self.color, self.line_width, self.resource_id)
        if len(self.points) < 3:
            msg = (
                "The region shape is a polygon. Polygons should have at least 3 points. "
                "If you would like to display a rectangle we recommend to use the designated rectangle shape."
            )
            emit_xmllib_input_warning(MessageInfo(msg, self.resource_id))

    def to_json_string(self) -> str:
        json_dict = {
            "status": "active" if self.active else "deleted",
            "type": "polygon",
            "lineWidth": self.line_width,
            "points": [x.to_dict() for x in self.points],
        }
        return json.dumps(json_dict)


@dataclass
class Circle(GeometryShape):
    center: GeometryPoint
    radius: Vector
    line_width: float
    color: str
    active: bool
    resource_id: str

    def __post_init__(self) -> None:
        _check_warn_shape_info(self.color, self.line_width, self.resource_id)

    def to_json_string(self) -> str:
        json_dict = {
            "status": "active" if self.active else "deleted",
            "type": "circle",
            "lineWidth": self.line_width,
            "points": [self.center.to_dict()],
            "radius": self.radius.to_dict(),
        }
        return json.dumps(json_dict)


@dataclass
class GeometryPoint:
    x: float
    y: float
    resource_id: str

    def __post_init__(self) -> None:
        if not all([is_decimal(self.x), is_decimal(self.y)]):
            msg = f"The entered geometry points are not floats. x: '{self.x}', y: '{self.y}'"
            emit_xmllib_input_warning(MessageInfo(msg, self.resource_id))
        else:
            info = []
            if not 0 <= float(self.x) <= 1:
                info.append(f"x: '{self.x}'")
            if not 0 <= float(self.y) <= 1:
                info.append(f"y: '{self.y}'")
            if info:
                msg = (
                    f"The geometry points must be larger/equal 0 and smaller/equal 1. "
                    f"The following points are not valid: {', '.join(info)}"
                )
                emit_xmllib_input_warning(MessageInfo(msg, self.resource_id))
            self.x = float(self.x)
            self.y = float(self.y)

    def to_dict(self) -> dict[str, float]:
        return {"x": self.x, "y": self.y}


@dataclass
class Vector:
    x: float
    y: float
    resource_id: str

    def __post_init__(self) -> None:
        if not all([is_decimal(self.x), is_decimal(self.y)]):
            msg = f"The radius vector are not floats. x: '{self.x}', y: '{self.y}'"
            emit_xmllib_input_warning(MessageInfo(msg, self.resource_id))
        else:
            info = []
            if not 0 <= float(self.x) <= 1:
                info.append(f"x: '{self.x}'")
            if not 0 <= float(self.y) <= 1:
                info.append(f"y: '{self.y}'")
            if info:
                msg = (
                    f"The radius vector must be larger/equal 0 and smaller/equal 1. "
                    f"The following values are not valid: {', '.join(info)}"
                )
                emit_xmllib_input_warning(MessageInfo(msg, self.resource_id))
            self.x = float(self.x)
            self.y = float(self.y)

    def to_dict(self) -> dict[str, float]:
        return {"x": self.x, "y": self.y}


def _check_warn_shape_info(color: str, line_width: float, res_id: str) -> None:
    problems = []
    line_msg = f"The line width must be a number between 1-5. The provided value is not valid: {line_width}"
    if not is_decimal(line_width):
        problems.append(line_msg)
    elif not 0 < int(line_width) <= 5:
        problems.append(line_msg)
    if not is_color(color):
        problems.append(f"The color must be a valid hexadecimal color. The provided value is not valid: {color}")
    if problems:
        msg = f"Some of the shape values are not valid: \n- {'\n- '.join(problems)}"
        emit_xmllib_input_warning(MessageInfo(msg, res_id))



import unittest
from typing import Any
from typing import cast

import pandas as pd
import pytest
import regex
from pandas.testing import assert_frame_equal

from dsp_tools.commands.excel2json import properties as e2j
from dsp_tools.commands.excel2json.models.input_error import InvalidExcelContentProblem
from dsp_tools.commands.excel2json.models.input_error import PropertyProblem
from dsp_tools.commands.excel2json.models.ontology import GuiAttributes
from dsp_tools.commands.excel2json.models.ontology import OntoProperty
from dsp_tools.error.exceptions import InputError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestFunctions(unittest.TestCase):
    def test_rename_deprecated_lang_cols(self) -> None:
        original_df = pd.DataFrame(
            {"en": [1, 2, 3], "de": [1, 2, 3], "fr": [1, 2, 3], "it": [1, 2, 3], "rm": [1, 2, 3]}
        )
        expected_df = pd.DataFrame(
            {
                "label_en": [1, 2, 3],
                "label_de": [1, 2, 3],
                "label_fr": [1, 2, 3],
                "label_it": [1, 2, 3],
                "label_rm": [1, 2, 3],
            }
        )
        with pytest.warns(UserWarning, match="column titles, which is deprecated"):
            returned_df = e2j._rename_deprecated_lang_cols(df=original_df)
            assert_frame_equal(expected_df, returned_df)
            returned_df = e2j._rename_deprecated_lang_cols(df=expected_df)
            assert_frame_equal(expected_df, returned_df)

    def test_do_property_excel_compliance_all_good(self) -> None:
        original_df = pd.DataFrame(
            {
                "name": ["name_1", "name_2", "name_3", "name_4", "name_5", "name_6"],
                "label_en": ["label_en_1", "label_en_2", pd.NA, pd.NA, "label_en_5", pd.NA],
                "label_de": ["label_de_1", pd.NA, "label_de_3", pd.NA, pd.NA, pd.NA],
                "label_fr": ["label_fr_1", pd.NA, pd.NA, "label_fr_4", pd.NA, pd.NA],
                "label_it": ["label_it_1", pd.NA, pd.NA, pd.NA, "label_it_5", pd.NA],
                "label_rm": ["label_rm_1", pd.NA, pd.NA, pd.NA, pd.NA, "label_rm_6"],
                "comment_en": ["comment_en_1", "comment_en_2", pd.NA, pd.NA, "comment_en_5", pd.NA],
                "comment_de": ["comment_de_1", pd.NA, "comment_de_3", pd.NA, pd.NA, pd.NA],
                "comment_fr": ["comment_fr_1", pd.NA, pd.NA, "comment_fr_4", pd.NA, pd.NA],
                "comment_it": ["comment_it_1", pd.NA, pd.NA, pd.NA, "comment_it_5", pd.NA],
                "comment_rm": ["comment_rm_1", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
                "super": ["super_1", "super_2", "super_3", "super_4.1, super_4.2, super_4.3", "super_5", "super_6"],
                "subject": ["subject_1", "subject_2", "subject_3", "subject_4", "subject_5", "subject_6"],
                "object": ["object_1", "object_2", "object_3", "object_4", "object_5", "object_6"],
                "gui_element": ["Simple", "Searchbox", "Date", "Searchbox", "List", "Searchbox"],
                "gui_attributes": ["size: 32, maxlength: 128", pd.NA, pd.NA, pd.NA, "hlist: languages", pd.NA],
            }
        )

        e2j._do_property_excel_compliance(df=original_df)

    def test_do_property_excel_compliance_problems(self) -> None:
        original_df = pd.DataFrame(
            {
                "name": ["name_1", "name_2", "name_3", "name_4", "name_5", "name_6", "name_7", pd.NA],
                "label_en": ["label_en_1", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
                "label_de": [pd.NA, pd.NA, "label_de_3", pd.NA, pd.NA, pd.NA, pd.NA, "label_de_8"],
                "label_fr": [pd.NA, pd.NA, pd.NA, "label_fr_4", pd.NA, pd.NA, pd.NA, pd.NA],
                "label_it": [pd.NA, pd.NA, pd.NA, pd.NA, "label_it_5", pd.NA, pd.NA, pd.NA],
                "label_rm": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "label_rm_6", pd.NA, pd.NA],
                "comment_en": ["comment_en_1", pd.NA, "comment_en_3", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
                "comment_de": [pd.NA, pd.NA, pd.NA, "comment_de_4", pd.NA, pd.NA, pd.NA, pd.NA],
                "comment_fr": [pd.NA, pd.NA, pd.NA, pd.NA, "comment_fr_5", pd.NA, pd.NA, pd.NA],
                "comment_it": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "comment_it_6", pd.NA, pd.NA],
                "comment_rm": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "comment_rm_7", pd.NA],
                "super": [
                    pd.NA,
                    "super_2",
                    pd.NA,
                    "super_4.1, super_4.2, super_4.3",
                    "super_5",
                    "super_6",
                    "super_7",
                    pd.NA,
                ],
                "subject": [
                    "subject_1",
                    "subject_2",
                    "subject_3",
                    "subject_4",
                    "subject_5",
                    "subject_6",
                    "subject_7",
                    pd.NA,
                ],
                "object": ["object_1", "object_2", "object_3", pd.NA, "object_5", "object_6", "object_7", pd.NA],
                "gui_element": ["Simple", "Searchbox", "Date", "Date", pd.NA, "List", pd.NA, pd.NA],
                "gui_attributes": ["size: 32, maxlength: 128", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
            }
        )
        expected_msg = regex.escape(
            "The Excel file 'properties.xlsx' contains the following problems:\n\n"
            "At the following locations, mandatory values are missing:\n"
            "    - Column 'name' | Row 9\n"
            "    - Column 'super' | Row 2\n"
            "    - Column 'super' | Row 4\n"
            "    - Column 'super' | Row 9\n"
            "    - Column 'object' | Row 5\n"
            "    - Column 'object' | Row 9\n"
            "    - Column 'gui_element' | Row 6\n"
            "    - Column 'gui_element' | Row 8\n"
            "    - Column 'gui_element' | Row 9\n"
            "    - Column 'label' | Row 3\n"
            "    - Column 'label' | Row 8\n"
            "    - Column 'gui_attributes' | Row 7"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j._do_property_excel_compliance(original_df)

    def test_rename_deprecated_hlist(self) -> None:
        original_df = pd.DataFrame({"hlist": [pd.NA, pd.NA, "languages"]})
        expected_df = pd.DataFrame({"gui_attributes": [pd.NA, pd.NA, "hlist:languages"]})
        with pytest.warns(UserWarning, match="'hlist', which is deprecated"):
            returned_df = e2j._rename_deprecated_hlist(df=original_df)
        assert_frame_equal(expected_df, returned_df)

        original_df = pd.DataFrame(
            {"hlist": [pd.NA, pd.NA, "languages"], "gui_attributes": [pd.NA, "attribute_1", pd.NA]}
        )
        expected_df = pd.DataFrame({"gui_attributes": [pd.NA, "attribute_1", "hlist:languages"]})
        with pytest.warns(UserWarning, match="'hlist', which is deprecated"):
            returned_df = e2j._rename_deprecated_hlist(df=original_df)
        assert_frame_equal(expected_df, returned_df)

    def test_unpack_gui_attributes(self) -> None:
        test_dict = {
            "maxlength:1, size:32": {"maxlength": "1", "size": "32"},
            "hlist: languages": {"hlist": "languages"},
            "hlist: special:character": {"hlist": "special:character"},
        }
        for original, expected in test_dict.items():
            self.assertDictEqual(e2j._unpack_gui_attributes(attribute_str=original), expected)

    def test_search_convert_numbers(self) -> None:
        test_dict = {"1": 1, "string": "string", "1.453": 1.453, "sdf.asdf": "sdf.asdf"}
        for original, expected in test_dict.items():
            self.assertEqual(e2j._search_convert_numbers_in_str(value_str=original), expected)

    def test_get_gui_attribute(self) -> None:
        original_df = pd.DataFrame(
            {"gui_attributes": [pd.NA, "max=1.4, min:1.2", "hlist:", "234345", "hlist: languages,"]}
        )
        self.assertIsNone(e2j._get_gui_attribute(df_row=cast("pd.Series[Any]", original_df.loc[0, :]), row_num=2))

        res_1 = e2j._get_gui_attribute(df_row=cast("pd.Series[Any]", original_df.loc[1, :]), row_num=3)
        assert isinstance(res_1, InvalidExcelContentProblem)
        assert res_1.excel_position.row == 3
        assert res_1.actual_content == "max=1.4, min:1.2"

        res_2 = e2j._get_gui_attribute(df_row=cast("pd.Series[Any]", original_df.loc[2, :]), row_num=4)
        assert isinstance(res_2, InvalidExcelContentProblem)
        assert res_2.excel_position.row == 4
        assert res_2.actual_content == "hlist:"

        res_3 = e2j._get_gui_attribute(df_row=cast("pd.Series[Any]", original_df.loc[3, :]), row_num=5)
        assert isinstance(res_3, InvalidExcelContentProblem)
        assert res_3.excel_position.row == 5
        assert res_3.actual_content == "234345"

        expected_dict = {"hlist": "languages"}
        returned_dict = e2j._get_gui_attribute(df_row=cast("pd.Series[Any]", original_df.loc[4, :]), row_num=6)
        assert isinstance(returned_dict, GuiAttributes)
        self.assertDictEqual(expected_dict, returned_dict.serialise())

    def test_check_compliance_gui_attributes_all_good(self) -> None:
        original_df = pd.DataFrame(
            {
                "gui_element": ["Spinbox", "List", "Searchbox", "Date", "Geonames", "Richtext", "TimeStamp"],
                "gui_attributes": ["Spinbox_attr", "List_attr", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
            }
        )
        assert not e2j._check_compliance_gui_attributes(df=original_df)

    def test_check_compliance_gui_attributes(self) -> None:
        original_df = pd.DataFrame(
            {
                "gui_element": ["Spinbox", "List", "Searchbox", "Date", "Geonames", "Richtext", "TimeStamp"],
                "gui_attributes": ["Spinbox_attr", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "TimeStamp_attr"],
            }
        )
        expected_dict = {"gui_attributes": [False, True, False, False, False, False, True]}
        returned_dict = e2j._check_compliance_gui_attributes(df=original_df)
        assert returned_dict
        casted_dict = {"gui_attributes": list(returned_dict["gui_attributes"])}
        self.assertDictEqual(expected_dict, casted_dict)

    def test_get_final_series_two_series(self) -> None:
        mandatory_check = pd.Series([True, True, False])
        no_attribute_check = pd.Series([True, False, False])
        expected = pd.Series([True, True, False])
        result = e2j._get_final_series(mandatory_check, no_attribute_check)
        self.assertEqual(list(result), list(expected))  # type: ignore[arg-type]

    def test_get_final_series_one_series_one_None(self) -> None:
        mandatory_check = pd.Series([True, True, False])
        no_attribute_check = None
        expected = pd.Series([True, True, False])
        result = e2j._get_final_series(mandatory_check, no_attribute_check)
        self.assertEqual(list(result), list(expected))  # type: ignore[arg-type]

    def test_get_final_series_one_None_one_series(self) -> None:
        mandatory_check = None
        no_attribute_check = pd.Series([True, False, False])
        expected = pd.Series([True, False, False])
        result = e2j._get_final_series(mandatory_check, no_attribute_check)
        self.assertEqual(list(result), list(expected))  # type: ignore[arg-type]

    def test_get_final_series_two_None(self) -> None:
        mandatory_check = None
        no_attribute_check = None
        expected = None
        result = e2j._get_final_series(mandatory_check, no_attribute_check)
        self.assertEqual(result, expected)

    def test_row2prop_duplicate_attrib_key_problem(self) -> None:
        test_series = pd.Series(
            {
                "name": "name_1",
                "label_en": "label_en",
                "label_de": "label_de",
                "label_fr": "label_fr",
                "label_it": "label_it",
                "label_rm": "label_rm",
                "comment_en": "comment_en",
                "comment_de": "comment_de",
                "comment_fr": "comment_fr",
                "comment_it": "comment_it",
                "comment_rm": "comment_rm",
                "super": "super_1",
                "subject": "subject_1",
                "object": "object_1",
                "gui_element": "Simple",
                "gui_attributes": "max:1.4 , max:1.2",
            }
        )
        res = e2j._row2prop(test_series, row_num=0)
        assert isinstance(res, PropertyProblem)
        assert res.prop_name == "name_1"
        assert len(res.problems) == 1
        invalid = res.problems[0]
        assert isinstance(invalid, InvalidExcelContentProblem)
        assert invalid.actual_content == "max:1.4 , max:1.2"
        assert invalid.expected_content == "attribute1: value, attribute2: value (no attribute key may be duplicated)"

    def test_row2prop(self) -> None:
        original_df = pd.DataFrame(
            {
                "name": ["name_1", "name_2", "name_3"],
                "label_en": ["label_en_1", "label_en_2", pd.NA],
                "label_de": ["label_de_1", pd.NA, "label_de_3"],
                "label_fr": ["label_fr_1", pd.NA, pd.NA],
                "label_it": ["label_it_1", pd.NA, pd.NA],
                "label_rm": ["label_rm_1", pd.NA, pd.NA],
                "comment_en": ["comment_en_1", "comment_en_2", pd.NA],
                "comment_de": ["comment_de_1", pd.NA, "comment_de_3"],
                "comment_fr": ["comment_fr_1", pd.NA, pd.NA],
                "comment_it": ["comment_it_1", pd.NA, pd.NA],
                "comment_rm": ["comment_rm_1", pd.NA, pd.NA],
                "super": ["super_1", "super_2.1, super_2.2", "super_3"],
                "subject": ["subject_1", "subject_2", pd.NA],
                "object": ["object_1", "object_2", "object_3"],
                "gui_element": ["Simple", "Date", "List"],
                "gui_attributes": ["size: 32, maxlength: 128", pd.NA, "hlist: Urheber:in"],
            }
        )
        returned_prop = e2j._row2prop(df_row=cast("pd.Series[Any]", original_df.loc[0, :]), row_num=0)
        expected_dict = {
            "name": "name_1",
            "object": "object_1",
            "subject": "subject_1",
            "gui_element": "Simple",
            "labels": {
                "en": "label_en_1",
                "de": "label_de_1",
                "fr": "label_fr_1",
                "it": "label_it_1",
                "rm": "label_rm_1",
            },
            "super": ["super_1"],
            "comments": {
                "en": "comment_en_1",
                "de": "comment_de_1",
                "fr": "comment_fr_1",
                "it": "comment_it_1",
                "rm": "comment_rm_1",
            },
            "gui_attributes": {"size": 32, "maxlength": 128},
        }
        assert isinstance(returned_prop, OntoProperty)
        self.assertDictEqual(expected_dict, returned_prop.serialise())

        returned_prop = e2j._row2prop(df_row=cast("pd.Series[Any]", original_df.loc[1, :]), row_num=1)
        expected_dict = {
            "comments": {"en": "comment_en_2"},
            "gui_element": "Date",
            "labels": {"en": "label_en_2"},
            "name": "name_2",
            "object": "object_2",
            "subject": "subject_2",
            "super": ["super_2.1", "super_2.2"],
        }
        assert isinstance(returned_prop, OntoProperty)
        self.assertDictEqual(expected_dict, returned_prop.serialise())

        returned_prop = e2j._row2prop(df_row=cast("pd.Series[Any]", original_df.loc[2, :]), row_num=2)
        expected_dict = {
            "comments": {"de": "comment_de_3"},
            "gui_attributes": {"hlist": "Urheber:in"},
            "gui_element": "List",
            "labels": {"de": "label_de_3"},
            "name": "name_3",
            "object": "object_3",
            "super": ["super_3"],
        }
        assert isinstance(returned_prop, OntoProperty)
        self.assertDictEqual(expected_dict, returned_prop.serialise())


@pytest.mark.parametrize(
    ("input_str", "expected_key", "expected_val"),
    [
        ("min:1.2", "min", "1.2"),
        ("hlist: Urheber:in", "hlist", "Urheber:in"),
        ("hlist:   Urheber : in", "hlist", "Urheber : in"),
        # While this does not make sense, it is not possible to allow ":"
        # in the text and catch these kinds of errors at the same time.
        ("max:1.4 / min:1.2", "max", "1.4 / min:1.2"),
    ],
)
def test_extract_information_from_single_gui_attribute_good(
    input_str: str, expected_key: str, expected_val: str
) -> None:
    attrib_key, attrib_val = e2j._extract_information_from_single_gui_attribute(input_str)
    assert attrib_key == expected_key
    assert attrib_val == expected_val


@pytest.mark.parametrize("input_str", ["hlist:", "234345"])
def test_extract_information_from_single_gui_attribute_raises(input_str: str) -> None:
    with pytest.raises(InputError):
        e2j._extract_information_from_single_gui_attribute(input_str)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

import regex
from loguru import logger

from dsp_tools.cli.args import ValidationSeverity


def _transform_server_url_to_foldername(server: str) -> str:
    """
    Take the server URL and transform it so that it can be used as foldername.

    Args:
        server: server, e.g. "https://api.test.dasch.swiss/" or "http://0.0.0.0:3333"

    Returns:
        simplified version, e.g. "test.dasch.swiss" or "localhost"
    """
    server_substitutions = {
        r"https?://": "",
        r"^api\.": "",
        r":\d{2,5}/?$": "",
        r"/$": "",
        r"0.0.0.0": "localhost",  # noqa: S104 (hardcoded-bind-all-interfaces)
    }
    for pattern, repl in server_substitutions.items():
        server = regex.sub(pattern, repl, server)
    return server


@dataclass(frozen=True)
class DiagnosticsConfig:
    """Configures all diagnostics for a given upload."""

    server_as_foldername: str = "unknown"
    save_location: Path = field(default=Path.home() / ".dsp-tools" / "xmluploads")


@dataclass(frozen=True)
class UploadConfig:
    """Configuration for the upload process."""

    media_previously_uploaded: bool = False
    server: str = "unknown"
    shortcode: str = "unknown"
    diagnostics: DiagnosticsConfig = field(default_factory=DiagnosticsConfig)
    interrupt_after: int | None = None
    skip_iiif_validation: bool = False
    validation_severity: ValidationSeverity = field(default_factory=lambda: ValidationSeverity.INFO)

    def with_server_info(
        self,
        server: str,
        shortcode: str,
    ) -> UploadConfig:
        """Create a new UploadConfig with the given server."""
        server_as_foldername = _transform_server_url_to_foldername(server)
        save_location = Path.home() / Path(".dsp-tools") / "xmluploads" / server_as_foldername / "resumable/latest.pkl"
        save_location.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"{save_location=:}")
        diagnostics: DiagnosticsConfig = dataclasses.replace(
            self.diagnostics,
            server_as_foldername=server_as_foldername,
            save_location=save_location,
        )
        return dataclasses.replace(self, server=server, shortcode=shortcode, diagnostics=diagnostics)



from rdflib import RDF
from rdflib import RDFS
from rdflib import XSD

from dsp_tools.commands.validate_data.constants import ARCHIVE_FILE_VALUE
from dsp_tools.commands.validate_data.constants import AUDIO_FILE_VALUE
from dsp_tools.commands.validate_data.constants import DOCUMENT_FILE_VALUE
from dsp_tools.commands.validate_data.constants import IIIF_URI_VALUE
from dsp_tools.commands.validate_data.constants import MOVING_IMAGE_FILE_VALUE
from dsp_tools.commands.validate_data.constants import STILL_IMAGE_FILE_VALUE
from dsp_tools.commands.validate_data.constants import TEXT_FILE_VALUE
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.rdf_like_data import TripleObjectType
from dsp_tools.commands.validate_data.models.rdf_like_data import TriplePropertyType
from dsp_tools.commands.validate_data.models.validation import ViolationType
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import BOOLEAN_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import COLOR_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import DECIMAL_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import GEOMETRY_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import GEONAME_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import INT_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import RICHTEXT_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import SIMPLE_TEXT_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import TIME_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.make_rdf_graph.constants import URI_PROP_TYPE_INFO
from dsp_tools.commands.xmlupload.models.rdf_models import RDFPropTypeInfo
from dsp_tools.utils.rdflib_constants import API_SHAPES
from dsp_tools.utils.rdflib_constants import KNORA_API
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType

FILE_TYPE_TO_PROP = {
    KnoraValueType.ARCHIVE_FILE: f"{KNORA_API_STR}hasArchiveFileValue",
    KnoraValueType.AUDIO_FILE: f"{KNORA_API_STR}hasAudioFileValue",
    KnoraValueType.DOCUMENT_FILE: f"{KNORA_API_STR}hasDocumentFileValue",
    KnoraValueType.MOVING_IMAGE_FILE: f"{KNORA_API_STR}hasMovingImageFileValue",
    KnoraValueType.STILL_IMAGE_FILE: f"{KNORA_API_STR}hasStillImageFileValue",
    KnoraValueType.STILL_IMAGE_IIIF: f"{KNORA_API_STR}hasStillImageFileValue",
    KnoraValueType.TEXT_FILE: f"{KNORA_API_STR}hasTextFileValue",
}

# Mappers from internal representation to API format
XML_TAG_TO_VALUE_TYPE_MAPPER = {
    "boolean-prop": KnoraValueType.BOOLEAN_VALUE,
    "color-prop": KnoraValueType.COLOR_VALUE,
    "date-prop": KnoraValueType.DATE_VALUE,
    "decimal-prop": KnoraValueType.DECIMAL_VALUE,
    "geometry-prop": KnoraValueType.GEOM_VALUE,
    "geoname-prop": KnoraValueType.GEONAME_VALUE,
    "list-prop": KnoraValueType.LIST_VALUE,
    "integer-prop": KnoraValueType.INT_VALUE,
    "resptr-prop": KnoraValueType.LINK_VALUE,
    "time-prop": KnoraValueType.TIME_VALUE,
    "uri-prop": KnoraValueType.URI_VALUE,
}

TRIPLE_PROP_TYPE_TO_IRI_MAPPER = {
    TriplePropertyType.RDF_TYPE: RDF.type,
    TriplePropertyType.RDFS_LABEL: RDFS.label,
    TriplePropertyType.KNORA_COMMENT_ON_VALUE: KNORA_API.valueHasComment,
    TriplePropertyType.KNORA_PERMISSIONS: KNORA_API.hasPermissions,
    TriplePropertyType.KNORA_INTERVAL_START: KNORA_API.intervalValueHasStart,
    TriplePropertyType.KNORA_INTERVAL_END: KNORA_API.intervalValueHasEnd,
    TriplePropertyType.KNORA_STANDOFF_LINK: KNORA_API.hasStandoffLinkTo,
    TriplePropertyType.KNORA_LICENSE: KNORA_API.hasLicense,
    TriplePropertyType.KNORA_AUTHORSHIP: KNORA_API.hasAuthorship,
    TriplePropertyType.KNORA_COPYRIGHT_HOLDER: KNORA_API.hasCopyrightHolder,
}

VALUE_INFO_TO_RDF_MAPPER = {
    KnoraValueType.BOOLEAN_VALUE: BOOLEAN_PROP_TYPE_INFO,
    KnoraValueType.COLOR_VALUE: COLOR_PROP_TYPE_INFO,
    KnoraValueType.DATE_VALUE: RDFPropTypeInfo(KNORA_API.DateValue, KNORA_API.valueAsString, XSD.string),
    KnoraValueType.DECIMAL_VALUE: DECIMAL_PROP_TYPE_INFO,
    KnoraValueType.GEONAME_VALUE: GEONAME_PROP_TYPE_INFO,
    KnoraValueType.GEOM_VALUE: GEOMETRY_PROP_TYPE_INFO,
    KnoraValueType.LIST_VALUE: RDFPropTypeInfo(KNORA_API.ListValue, KNORA_API.listValueAsListNode, XSD.string),
    KnoraValueType.LINK_VALUE: RDFPropTypeInfo(KNORA_API.LinkValue, API_SHAPES.linkValueHasTargetID, XSD.string),
    KnoraValueType.INT_VALUE: INT_PROP_TYPE_INFO,
    KnoraValueType.INTERVAL_VALUE: RDFPropTypeInfo(KNORA_API.IntervalValue, KNORA_API.hasSegmentBounds),
    KnoraValueType.SIMPLETEXT_VALUE: SIMPLE_TEXT_PROP_TYPE_INFO,
    KnoraValueType.RICHTEXT_VALUE: RICHTEXT_PROP_TYPE_INFO,
    KnoraValueType.TIME_VALUE: TIME_PROP_TYPE_INFO,
    KnoraValueType.URI_VALUE: URI_PROP_TYPE_INFO,
    KnoraValueType.ARCHIVE_FILE: ARCHIVE_FILE_VALUE,
    KnoraValueType.AUDIO_FILE: AUDIO_FILE_VALUE,
    KnoraValueType.DOCUMENT_FILE: DOCUMENT_FILE_VALUE,
    KnoraValueType.MOVING_IMAGE_FILE: MOVING_IMAGE_FILE_VALUE,
    KnoraValueType.STILL_IMAGE_FILE: STILL_IMAGE_FILE_VALUE,
    KnoraValueType.STILL_IMAGE_IIIF: IIIF_URI_VALUE,
    KnoraValueType.TEXT_FILE: TEXT_FILE_VALUE,
}

VALUE_INFO_TRIPLE_OBJECT_TYPE = {
    KnoraValueType.BOOLEAN_VALUE: TripleObjectType.BOOLEAN,
    KnoraValueType.COLOR_VALUE: TripleObjectType.STRING,
    KnoraValueType.DATE_VALUE: TripleObjectType.STRING,
    KnoraValueType.DECIMAL_VALUE: TripleObjectType.DECIMAL,
    KnoraValueType.GEOM_VALUE: TripleObjectType.STRING,
    KnoraValueType.GEONAME_VALUE: TripleObjectType.STRING,
    KnoraValueType.LIST_VALUE: TripleObjectType.STRING,
    KnoraValueType.LINK_VALUE: TripleObjectType.IRI,
    KnoraValueType.INT_VALUE: TripleObjectType.INTEGER,
    KnoraValueType.SIMPLETEXT_VALUE: TripleObjectType.STRING,
    KnoraValueType.RICHTEXT_VALUE: TripleObjectType.STRING,
    KnoraValueType.TIME_VALUE: TripleObjectType.DATETIME,
    KnoraValueType.URI_VALUE: TripleObjectType.URI,
    KnoraValueType.ARCHIVE_FILE: TripleObjectType.STRING,
    KnoraValueType.AUDIO_FILE: TripleObjectType.STRING,
    KnoraValueType.DOCUMENT_FILE: TripleObjectType.STRING,
    KnoraValueType.MOVING_IMAGE_FILE: TripleObjectType.STRING,
    KnoraValueType.STILL_IMAGE_FILE: TripleObjectType.STRING,
    KnoraValueType.STILL_IMAGE_IIIF: TripleObjectType.URI,
    KnoraValueType.TEXT_FILE: TripleObjectType.STRING,
}

TRIPLE_OBJECT_TYPE_TO_XSD = {
    TripleObjectType.BOOLEAN: XSD.boolean,
    TripleObjectType.DATETIME: XSD.dateTimeStamp,
    TripleObjectType.DECIMAL: XSD.decimal,
    TripleObjectType.INTEGER: XSD.integer,
    TripleObjectType.STRING: XSD.string,
    TripleObjectType.URI: XSD.anyURI,
}

# validation results

RESULT_TO_PROBLEM_MAPPER = {
    ViolationType.SEQNUM_IS_PART_OF: ProblemType.GENERIC,
    ViolationType.UNIQUE_VALUE: ProblemType.DUPLICATE_VALUE,
    ViolationType.VALUE_TYPE: ProblemType.VALUE_TYPE_MISMATCH,
    ViolationType.PATTERN: ProblemType.INPUT_REGEX,
    ViolationType.GENERIC: ProblemType.GENERIC,
    ViolationType.LINK_TARGET: ProblemType.LINK_TARGET_TYPE_MISMATCH,
    ViolationType.MAX_CARD: ProblemType.MAX_CARD,
    ViolationType.MIN_CARD: ProblemType.MIN_CARD,
    ViolationType.NON_EXISTING_CARD: ProblemType.NON_EXISTING_CARD,
    ViolationType.FILEVALUE_PROHIBITED: ProblemType.FILE_VALUE_PROHIBITED,
    ViolationType.FILE_VALUE: ProblemType.FILE_VALUE,
    ViolationType.FILE_DUPLICATE: ProblemType.FILE_DUPLICATE,
}



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest

from dsp_tools.commands.validate_data.get_user_validation_message import _get_message_for_one_resource
from dsp_tools.commands.validate_data.get_user_validation_message import _shorten_input
from dsp_tools.commands.validate_data.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.models.input_problems import AllProblems
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent


@pytest.fixture
def generic_problem() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.GENERIC,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasGenericProblem",
        severity=Severity.VIOLATION,
        message="This is a generic problem.",
    )


@pytest.fixture
def file_value() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.FILE_VALUE,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="bitstream / iiif-uri",
        severity=Severity.VIOLATION,
        expected="A MovingImageRepresentation requires a file with the extension 'mp4'.",
    )


@pytest.fixture
def max_card() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.MAX_CARD,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasMaxCardProblem",
        severity=Severity.VIOLATION,
        expected="Cardinality 1",
    )


@pytest.fixture
def min_card() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.MIN_CARD,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasMinCardProblem",
        severity=Severity.VIOLATION,
        expected="Cardinality 1-n",
    )


@pytest.fixture
def non_existing_card() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.NON_EXISTING_CARD,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
    )


@pytest.fixture
def file_value_prohibited() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.FILE_VALUE_PROHIBITED,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="bitstream / iiif-uri",
        severity=Severity.VIOLATION,
    )


@pytest.fixture
def link_value_type_mismatch() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        input_type="LinkValue",
        expected="ListValue",
    )


@pytest.fixture
def input_regex() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.INPUT_REGEX,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        input_value="wrong input",
        expected="Expected format information",
    )


@pytest.fixture
def link_target_mismatch() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.LINK_TARGET_TYPE_MISMATCH,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        input_value="link_target_id",
        input_type="onto:Class",
        expected="onto:File or a subclass",
    )


@pytest.fixture
def inexistent_linked_resource() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.INEXISTENT_LINKED_RESOURCE,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        input_value="link_target_id",
    )


@pytest.fixture
def duplicate_value() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.DUPLICATE_VALUE,
        res_id="res_id",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        input_value="Text",
    )


@pytest.fixture
def missing_legal_warning() -> InputProblem:
    return InputProblem(
        problem_type=ProblemType.GENERIC,
        res_id="image_no_legal_info",
        res_type="onto:TestStillImageRepresentation",
        prop_name="bitstream / iiif-uri",
        severity=Severity.WARNING,
        expected="Files and IIIF-URIs require a reference to a license.",
    )


def test_sort_user_problems_with_iris(duplicate_value, link_value_type_mismatch, missing_legal_warning):
    references_iri = InputProblem(
        problem_type=ProblemType.INEXISTENT_LINKED_RESOURCE,
        res_id="references_iri",
        res_type="onto:Class",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        input_value="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
    )
    inexistent_license_iri = InputProblem(
        problem_type=ProblemType.GENERIC,
        res_id="inexistent_license_iri",
        res_type="onto:TestStillImageRepresentation",
        prop_name="bitstream / iiif-uri",
        severity=Severity.VIOLATION,
        input_value="http://rdfh.ch/licenses/this-iri-does-not-exist",
        message="Files and IIIF-URIs require a reference to a license.",
    )
    result = sort_user_problems(
        AllProblems(
            [duplicate_value, link_value_type_mismatch, references_iri, inexistent_license_iri, missing_legal_warning],
            [],
        )
    )
    assert len(result.unique_violations) == 3
    assert set([x.res_id for x in result.unique_violations]) == {"res_id", "inexistent_license_iri"}
    assert len(result.user_warnings) == 1
    assert result.user_warnings[0].res_id == "image_no_legal_info"
    assert len(result.user_info) == 1
    assert result.user_info[0].res_id == "references_iri"
    assert not result.unexpected_shacl_validation_components


def test_sort_user_problems_with_duplicate(duplicate_value, link_value_type_mismatch):
    duplicate_message_should_stay = InputProblem(
        problem_type=ProblemType.FILE_DUPLICATE,
        res_id="file_value_duplicate",
        res_type="onto:TestStillImageRepresentation",
        prop_name="bitstream / iiif-uri",
        severity=Severity.INFO,
        input_value="duplicate_file.zip",
        expected="The entered filepath is used more than once in your data.",
    )
    duplicate_message_should_be_removed = InputProblem(
        problem_type=ProblemType.FILE_DUPLICATE,
        res_id="file_value_duplicate",
        res_type="onto:TestStillImageRepresentation",
        prop_name="bitstream / iiif-uri",
        severity=Severity.INFO,
        input_value="duplicate_file.zip",
        expected="The entered filepath is used more than once in your data.",
    )
    file_duplicate_resource_other_problem_should_stay = InputProblem(
        problem_type=ProblemType.GENERIC,
        res_id="file_value_duplicate",
        res_type="",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        expected="Some Expectation",
    )
    should_remain = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="text_value_id",
        res_type="",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        expected="TextValue without formatting",
    )
    should_be_removed = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="text_value_id",
        res_type="",
        prop_name="onto:hasProp",
        severity=Severity.VIOLATION,
        expected="This property requires a TextValue",
    )
    result = sort_user_problems(
        AllProblems(
            [
                duplicate_value,
                link_value_type_mismatch,
                should_remain,
                should_be_removed,
                duplicate_message_should_stay,
                duplicate_message_should_be_removed,
                file_duplicate_resource_other_problem_should_stay,
            ],
            [UnexpectedComponent("sh:unexpected"), UnexpectedComponent("sh:unexpected")],
        )
    )
    assert len(result.unique_violations) == 4
    assert len(result.user_info) == 1
    assert result.user_info[0].res_id == "file_value_duplicate"
    assert len(result.unexpected_shacl_validation_components) == 1
    assert set([x.res_id for x in result.unique_violations]) == {"text_value_id", "res_id", "file_value_duplicate"}


def test_sort_user_problems_different_props():
    one = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="res_id",
        res_type="",
        prop_name="onto:prop2",
        severity=Severity.VIOLATION,
        expected="TextValue without formatting",
    )
    two = InputProblem(
        problem_type=ProblemType.VALUE_TYPE_MISMATCH,
        res_id="res_id",
        res_type="",
        prop_name="onto:prop1",
        severity=Severity.VIOLATION,
        expected="This property requires a TextValue",
    )
    result = sort_user_problems(AllProblems([one, two], []))
    assert len(result.unique_violations) == 2
    assert not result.user_info
    assert not result.unexpected_shacl_validation_components
    assert [x.res_id for x in result.unique_violations] == ["res_id", "res_id"]


def test_get_message_for_one_resource_generic(generic_problem):
    result = _get_message_for_one_resource([generic_problem])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\nonto:hasGenericProblem\n    - This is a generic problem."
    )
    assert result == expected


def test_get_message_for_one_resource_file_value(file_value):
    result = _get_message_for_one_resource([file_value])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "bitstream / iiif-uri\n"
        "    - Expected: A MovingImageRepresentation requires a file with the extension 'mp4'."
    )
    assert result == expected


def test_get_message_for_one_resource_max_card(max_card):
    result = _get_message_for_one_resource([max_card])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasMaxCardProblem\n"
        "    - Maximum Cardinality Violation | Expected: Cardinality 1"
    )
    assert result == expected


def test_get_message_for_one_resource_min_card(min_card):
    result = _get_message_for_one_resource([min_card])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasMinCardProblem\n"
        "    - Minimum Cardinality Violation | Expected: Cardinality 1-n"
    )
    assert result == expected


def test_get_message_for_one_resource_non_existing_card(non_existing_card):
    result = _get_message_for_one_resource([non_existing_card])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - The resource class does not have a cardinality for this property."
    )
    assert result == expected


def test_get_message_for_one_resource_file_value_prohibited(file_value_prohibited):
    result = _get_message_for_one_resource([file_value_prohibited])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "bitstream / iiif-uri\n"
        "    - A file was added to the resource. This resource type must not have a file."
    )
    assert result == expected


def test_get_message_for_one_resource_value_type_mismatch(link_value_type_mismatch):
    result = _get_message_for_one_resource([link_value_type_mismatch])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Value Type Mismatch | Actual input type: 'LinkValue' | Expected Value Type: ListValue"
    )
    assert result == expected


def test_get_message_for_one_resource_input_regex(input_regex):
    result = _get_message_for_one_resource([input_regex])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Wrong Format of Input | Your input: 'wrong input' | "
        "Expected Input Format: Expected format information"
    )
    assert result == expected


def test_get_message_for_one_resource_link_target_mismatch(link_target_mismatch):
    result = _get_message_for_one_resource([link_target_mismatch])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Linked Resource Type Mismatch | Your input: 'link_target_id' | "
        "Actual input type: 'onto:Class' | Expected Resource Type: onto:File or a subclass"
    )
    assert result == expected


def test_get_message_for_one_resource_inexistent_linked_resource(inexistent_linked_resource):
    result = _get_message_for_one_resource([inexistent_linked_resource])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Linked Resource does not exist | Your input: 'link_target_id'"
    )
    assert result == expected


def test_get_message_for_one_resource_duplicate_value(duplicate_value):
    result = _get_message_for_one_resource([duplicate_value])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "onto:hasProp\n"
        "    - Your input is duplicated | Your input: 'Text'"
    )
    assert result == expected


def test_get_message_for_one_resource_several_problems(file_value, inexistent_linked_resource):
    result = _get_message_for_one_resource([file_value, inexistent_linked_resource])
    expected = (
        "Resource ID: res_id | Resource Type: onto:Class\n"
        "bitstream / iiif-uri\n"
        "    - Expected: A MovingImageRepresentation requires a file with the extension 'mp4'.\n"
        "onto:hasProp\n"
        "    - Linked Resource does not exist | Your input: 'link_target_id'"
    )
    assert result == expected


@pytest.mark.parametrize(
    ("user_input", "problem_type", "expected"),
    [
        (
            "this/is/a/very/very/very/long/filepath/file.csv",
            ProblemType.FILE_VALUE,
            "this/is/a/very/very/very/long/filepath/file.csv",
        ),
        (
            "This is a very long sentence and should be shortened.",
            ProblemType.GENERIC,
            "This is a very long sentence and should [...]",
        ),
        ("So short, nothing happens", ProblemType.VALUE_TYPE_MISMATCH, "So short, nothing happens"),
        (None, ProblemType.GENERIC, None),
    ],
)
def test_shorten_input(user_input, problem_type, expected):
    result = _shorten_input(user_input, problem_type)
    assert result == expected



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from typing import Any

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.error.xmllib_warnings_util import raise_input_error
from dsp_tools.xmllib.internal.checkers import check_and_warn_potentially_empty_string
from dsp_tools.xmllib.internal.input_converters import check_and_fix_collection_input
from dsp_tools.xmllib.models.config_options import NewlineReplacement
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.internal.geometry import Circle
from dsp_tools.xmllib.models.internal.geometry import GeometryPoint
from dsp_tools.xmllib.models.internal.geometry import GeometryShape
from dsp_tools.xmllib.models.internal.geometry import Polygon
from dsp_tools.xmllib.models.internal.geometry import Rectangle
from dsp_tools.xmllib.models.internal.geometry import Vector
from dsp_tools.xmllib.models.internal.migration_metadata import MigrationMetadata
from dsp_tools.xmllib.models.internal.values import LinkValue
from dsp_tools.xmllib.models.internal.values import Richtext
from dsp_tools.xmllib.models.internal.values import SimpleText
from dsp_tools.xmllib.models.internal.values import Value
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_nonempty_value

# ruff: noqa: D101, D102

LIST_SEPARATOR = "\n    - "


@dataclass
class RegionResource:
    res_id: str
    label: str
    values: list[Value]
    geometry: GeometryShape | None
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        check_and_warn_potentially_empty_string(value=self.label, res_id=self.res_id, expected="string", field="label")

    @staticmethod
    def create_new(
        res_id: str,
        label: str,
        region_of: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    ) -> RegionResource:
        """
        Creates a new region resource.
        A region is a region of interest (ROI) in a StillImageRepresentation.

        Exactly one geometry shape has to be added to the resource with one of the following methods:
        `add_rectangle`, `add_polygon`, `add_circle` (see documentation below for more information).

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#region)

        Args:
            res_id: ID of this region resource
            label: label of this region resource
            region_of: ID of the image resource that this region refers to (cardinality 1)
            permissions: permissions of this region resource

        Returns:
            A region resource

        Examples:
            ```python
            region = xmllib.RegionResource.create_new(
                res_id="ID",
                label="label",
                region_of="image_id",
            )
            ```
        """
        return RegionResource(
            res_id=res_id,
            label=label,
            values=[
                LinkValue.new(
                    value=region_of, prop_name="isRegionOf", resource_id=res_id, comment=None, permissions=permissions
                )
            ],
            geometry=None,
            permissions=permissions,
        )

    def add_rectangle(
        self,
        point1: tuple[float, float],
        point2: tuple[float, float],
        line_width: float = 2,
        color: str = "#5b24bf",
        active: bool = True,
    ) -> RegionResource:
        """
        Add a rectangle shape to the region.

        [For a visual example see the XML documentation](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geometry)

        Args:
            point1: first point of the rectangle represented as two numbers between 0 and 1 in the format (x, y)
            point2: second point of the rectangle represented as two numbers between 0 and 1 in the format (x, y)
            line_width: A number in pixels between 1 - 5
            color: A hexadecimal color value which starts with a `#` followed by 3 or 6 numerals.
                The default value was chosen as it is distinguishable for most color-blind people.
            active: If set to `False`, the region is marked as 'deleted'

        Returns:
            Region with added rectangle

        Examples:
            ```python
            region = region.add_rectangle(
                point1=(0.1, 0.1),
                point2=(0.2, 0.2),
            )
            ```

            ```python
            # with custom display values
            region = region.add_rectangle(
                point1=(0.1, 0.1),
                point2=(0.2, 0.2),
                line_width=3,
                color="#32a873",
            )
            ```
        """
        self.geometry = Rectangle(
            point1=GeometryPoint(point1[0], point1[1], self.res_id),
            point2=GeometryPoint(point2[0], point2[1], self.res_id),
            line_width=line_width,
            color=color,
            active=active,
            resource_id=self.res_id,
        )
        return self

    def add_polygon(
        self,
        points: list[tuple[float, float]],
        line_width: float = 2,
        color: str = "#5b24bf",
        active: bool = True,
    ) -> RegionResource:
        """
        Add a polygon shape to the region.
        A polygon should have at least three points.
        If you wish to create a rectangle, please use the designated `add_rectangle` method.

        [For a visual example see the XML documentation](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geometry)

        **Please note that this cannot currently be displayed in the dsp-app.**

        Args:
            points: list of tuples containing two numbers between 0 and 1 in the format (x, y)
            line_width: A number in pixels between 1 - 5
            color: A hexadecimal color value which starts with a `#` followed by 3 or 6 numerals.
                The default value was chosen as it is distinguishable for most color-blind people.
            active: If set to `False` the region is marked as 'deleted'

        Returns:
            Region with added polygon

        Examples:
            ```python
            region = region.add_polygon(
                points=[(0.1, 0.1), (0.2, 0.2), (0.3, 0.3)],
            )
            ```

            ```python
            # with custom display values
            region = region.add_polygon(
                points=[(0.1, 0.1), (0.2, 0.2), (0.3, 0.3)],
                line_width=3,
                color="#32a873",
            )
            ```
        """
        geom_points = [GeometryPoint(val[0], val[1], self.res_id) for val in points]
        self.geometry = Polygon(
            points=geom_points, line_width=line_width, color=color, active=active, resource_id=self.res_id
        )
        return self

    def add_circle(
        self,
        center: tuple[float, float],
        radius: tuple[float, float],
        line_width: float = 2,
        color: str = "#5b24bf",
        active: bool = True,
    ) -> RegionResource:
        """
        Add a circle shape to the region.

        [For a visual example see the XML documentation](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geometry)

        **Please note that this cannot currently be displayed in the dsp-app.**

        Args:
            center: center of the circle, represented as two numbers between 0 and 1 in the format (x, y)
            radius: radius of the circle, represented as a 2-dimensional vector,
                i.e. two numbers between 0 and 1 in the format (x, y)
            line_width: A number in pixels between 1 - 5
            color: A hexadecimal color value which starts with a `#` followed by 3 or 6 numerals.
                The default value was chosen as it is distinguishable for most color-blind people.
            active: If set to `False` the region is marked as 'deleted'

        Returns:
            Region with added circle

        Examples:
            ```python
            region = region.add_circle(
                center=(0.1, 0.1),
                radius=(0.2, 0.2),
            )
            ```

            ```python
            # with custom display values
            region = region.add_circle(
                center=(0.1, 0.1),
                radius=(0.2, 0.2),
                line_width=3,
                color="#32a873",
            )
            ```
        """
        self.geometry = Circle(
            center=GeometryPoint(center[0], center[1], self.res_id),
            radius=Vector(radius[0], radius[1], self.res_id),
            line_width=line_width,
            color=color,
            active=active,
            resource_id=self.res_id,
        )
        return self

    def add_comment(
        self,
        text: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> RegionResource:
        """
        Add a comment to the region

        Args:
            text: text of the comment
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original region, with the added comment

        Examples:
            ```python
            region = region.add_comment("comment text")
            ```

            ```python
            region = region.add_comment(text="comment text", comment="Comment about the comment.")
            ```
        """
        self.values.append(
            Richtext.new(
                value=text,
                prop_name="hasComment",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_comment_multiple(
        self,
        texts: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> RegionResource:
        """
        Add several comments to the region

        Args:
            texts: list of texts
            permissions: optional permissions of these values
            comment: optional comment about these comments
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original region, with the added comments

        Examples:
            ```python
            region = region.add_comment_multiple(["comment 1", "comment 2"])
            ```
        """
        vals = check_and_fix_collection_input(texts, "hasComment", self.res_id)
        comnts = [
            Richtext.new(
                value=x,
                prop_name="hasComment",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
            for x in vals
        ]
        self.values.extend(comnts)
        return self

    def add_comment_optional(
        self,
        text: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> RegionResource:
        """
        If the value is not empty, add it as comment, otherwise return the region unchanged.

        Args:
            text: text of the comment (or empty value)
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original region, with the added comment

        Examples:
            ```python
            region = region.add_comment_optional("comment text")
            ```

            ```python
            region = region.add_comment_optional(None)
            ```
        """
        if is_nonempty_value(text):
            return self.add_comment(text, permissions, comment, newline_replacement)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> RegionResource:
        """
        Add metadata from a SALSAH migration.

        Args:
            creation_date: creation date of the region in SALSAH
            iri: Original IRI in SALSAH
            ark: Original ARK in SALSAH

        Raises:
            InputError: if metadata already exists

        Returns:
            The original region, with the added metadata

        Examples:
            ```python
            region = region.add_migration_metadata(
                iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
                creation_date="1999-12-31T23:59:59.9999999+01:00"
            )
            ```
        """
        if self.migration_metadata:
            msg_info = MessageInfo(
                "This resource already contains migration metadata, no new data can be added.", resource_id=self.res_id
            )
            raise_input_error(msg_info)
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self


@dataclass
class LinkResource:
    res_id: str
    label: str
    values: list[Value]
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        check_and_warn_potentially_empty_string(value=self.label, res_id=self.res_id, expected="string", field="label")

    @staticmethod
    def create_new(
        res_id: str,
        label: str,
        link_to: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    ) -> LinkResource:
        """
        Creates a new link resource.
        A link resource is like a container that groups together several other resources of different classes.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#link)

        Args:
            res_id: ID of this link resource
            label: label of this link resource
            link_to: IDs of the resources that should be linked together (cardinality 1-n)
            permissions: permissions of this link resource

        Returns:
            A link resource

        Examples:
            ```python
            link_resource = xmllib.LinkResource.create_new(
                res_id="ID",
                label="label",
                link_to=["target_resource_id_1", "target_resource_id_2"],
            )
            ```
        """
        links_to = check_and_fix_collection_input(link_to, "hasLinkTo", res_id)
        link_vals: list[Value] = [
            LinkValue.new(value=x, prop_name="hasLinkTo", resource_id=res_id, comment=None, permissions=permissions)
            for x in links_to
        ]
        return LinkResource(
            res_id=res_id,
            label=label,
            values=link_vals,
            permissions=permissions,
        )

    def add_comment(
        self,
        text: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> LinkResource:
        """
        Add a comment to the resource

        Args:
            text: text of the comment
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comment

        Examples:
            ```python
            link_resource = link_resource.add_comment("comment text")
            ```

            ```python
            link_resource = link_resource.add_comment(text="comment text", comment="Comment about the comment.")
            ```
        """
        self.values.append(
            Richtext.new(
                value=text,
                prop_name="hasComment",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_comment_multiple(
        self,
        texts: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> LinkResource:
        """
        Add several comments to the resource

        Args:
            texts: list of texts
            permissions: optional permissions of these values
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comments

        Examples:
            ```python
            link_resource = link_resource.add_comment_multiple(["comment 1", "comment 2"])
            ```
        """
        vals = check_and_fix_collection_input(texts, "hasComment", self.res_id)
        for v in vals:
            self.add_comment(v, permissions, comment, newline_replacement)
        return self

    def add_comment_optional(
        self,
        text: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> LinkResource:
        """
        If the value is not empty, add it as comment, otherwise return the resource unchanged.

        Args:
            text: text of the comment (or empty value)
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comment

        Examples:
            ```python
            link_resource = link_resource.add_comment_optional("comment text")
            ```

            ```python
            link_resource = link_resource.add_comment_optional(None)
            ```
        """
        if is_nonempty_value(text):
            return self.add_comment(text, permissions, comment, newline_replacement)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> LinkResource:
        """
        Add metadata from a SALSAH migration.

        Args:
            creation_date: creation date of the resource in SALSAH
            iri: Original IRI in SALSAH
            ark: Original ARK in SALSAH

        Raises:
            InputError: if metadata already exists

        Returns:
            The original resource, with the added metadata

        Examples:
            ```python
            link_resource = link_resource.add_migration_metadata(
                iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
                creation_date="1999-12-31T23:59:59.9999999+01:00"
            )
            ```
        """
        if self.migration_metadata:
            msg_info = MessageInfo(
                "This resource already contains migration metadata, no new data can be added.", resource_id=self.res_id
            )
            raise_input_error(msg_info)
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self


@dataclass
class SegmentBounds:
    segment_start: float | int | str
    segment_end: float | int | str
    permissions: Permissions
    res_id: str

    def __post_init__(self) -> None:
        msg: list[str] = []
        if not is_decimal(self.segment_start):
            msg.append(f"Segment Start Value: {self.segment_start} | Type: {type(self.segment_start)}")
        if not is_decimal(self.segment_end):
            msg.append(f"Segment End Value: {self.segment_end} | Type: {type(self.segment_start)}")
        if msg:
            wrng = f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(msg)}"
            msg_info = MessageInfo(
                f"Segment bounds must be a float or integer. The following places have an unexpected type: {wrng}",
                self.res_id,
            )
            emit_xmllib_input_warning(msg_info)


@dataclass
class VideoSegmentResource:
    res_id: str
    label: str
    segment_bounds: SegmentBounds
    values: list[Value]
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        check_and_warn_potentially_empty_string(value=self.label, res_id=self.res_id, expected="string", field="label")

    @staticmethod
    def create_new(
        res_id: str,
        label: str,
        segment_of: str,
        segment_start: float | int | str,
        segment_end: float | int | str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    ) -> VideoSegmentResource:
        """
        Creates a new video segment resource, i.e. a time span of a MovingImageRepresentation.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#video-segment-and-audio-segment)

        Args:
            res_id: ID of this video segment resource
            label: label of this video segment resource
            segment_of: ID of the video resource that this segment refers to (cardinality 1)
            segment_start: start of the segment in seconds (cardinality 1)
            segment_end: end of the segment in seconds (cardinality 1)
            permissions: permissions of this resource

        Returns:
            A video segment resource

        Examples:
            ```python
            video_segment = xmllib.VideoSegmentResource.create_new(
                res_id="ID",
                label="label",
                segment_of="video_resource_id",
                segment_start=1.2,
                segment_end=3.4,
            )
            ```
        """
        segment_of_val = LinkValue.new(
            value=segment_of, prop_name="isSegmentOf", permissions=permissions, comment=None, resource_id=res_id
        )
        return VideoSegmentResource(
            res_id=res_id,
            label=label,
            values=[segment_of_val],
            segment_bounds=SegmentBounds(segment_start, segment_end, permissions, res_id),
            permissions=permissions,
        )

    def add_title(
        self,
        title: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        Add a title to the resource.

        Args:
            title: text
            permissions: permissions of the value
            comment: comments on the value

        Returns:
            The original resource, with the added title

        Examples:
            ```python
            video_segment = video_segment.add_title("segment title")
            ```
        """
        existing_titles = [x for x in self.values if x.prop_name == "hasTitle"]
        if any(existing_titles):
            _warn_value_exists(
                old_value=existing_titles.pop(0).value, new_value=title, value_field="title", res_id=self.res_id
            )
        self.values.append(
            SimpleText.new(
                value=title,
                prop_name="hasTitle",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_title_optional(
        self,
        title: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        If the value is not empty, add it as title, otherwise return the resource unchanged.

        Args:
            title: text or empty value
            permissions: permissions of the value
            comment: comments on the value

        Returns:
            The original resource, with the added title

        Examples:
            ```python
            video_segment = video_segment.add_title("segment title")
            ```

            ```python
            video_segment = video_segment.add_title(None)
            ```
        """
        if is_nonempty_value(title):
            self.add_title(title, permissions, comment)
        return self

    def add_comment(
        self,
        text: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> VideoSegmentResource:
        """
        Add a comment to the resource

        Args:
            text: text
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comment

        Examples:
            ```python
            video_segment = video_segment.add_comment("comment text")
            ```
        """
        self.values.append(
            Richtext.new(
                value=text,
                prop_name="hasComment",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_comment_multiple(
        self,
        texts: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> VideoSegmentResource:
        """
        Add several comments to the resource

        Args:
            texts: list of texts
            permissions: optional permissions of these values
            comment: optional comment about these comments
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comments

        Examples:
            ```python
            video_segment = video_segment.add_comment_multiple(["comment 1", "comment 2"])
            ```
        """
        vals = check_and_fix_collection_input(texts, "hasComment", self.res_id)
        for v in vals:
            self.add_comment(v, permissions, comment, newline_replacement)
        return self

    def add_comment_optional(
        self,
        text: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> VideoSegmentResource:
        """
        If the value is not empty, add it as comment, otherwise return the resource unchanged.

        Args:
            text: text of the comment (or empty value)
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comment

        Examples:
            ```python
            video_segment = video_segment.add_comment_optional("comment text")
            ```

            ```python
            video_segment = video_segment.add_comment_optional(None)
            ```
        """
        if is_nonempty_value(text):
            self.add_comment(text, permissions, comment, newline_replacement)
        return self

    def add_description(
        self,
        description: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> VideoSegmentResource:
        """
        Add a description to the resource

        Args:
            description: text
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added description

        Examples:
            ```python
            video_segment = video_segment.add_description("description text")
            ```
        """
        self.values.append(
            Richtext.new(
                value=description,
                prop_name="hasDescription",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_description_multiple(
        self,
        descriptions: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> VideoSegmentResource:
        """
        Add several descriptions to the resource

        Args:
            descriptions: list of texts
            permissions: optional permissions of these value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added descriptions

        Examples:
            ```python
            video_segment = video_segment.add_description_multiple(["description 1", "description 2"])
            ```
        """
        vals = check_and_fix_collection_input(descriptions, "description", self.res_id)
        for v in vals:
            self.add_description(v, permissions, comment, newline_replacement)
        return self

    def add_description_optional(
        self,
        description: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> VideoSegmentResource:
        """
        If the value is not empty, add it as description, otherwise return the resource unchanged.

        Args:
            description: text or empty value
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added description

        Examples:
            ```python
            video_segment = video_segment.add_description_optional("description text")
            ```

            ```python
            video_segment = video_segment.add_description_optional(None)
            ```
        """
        if is_nonempty_value(description):
            self.add_description(description, permissions, comment, newline_replacement)
        return self

    def add_keyword(
        self,
        keyword: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        Add a keyword to the resource

        Args:
            keyword: text
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added keyword

        Examples:
            ```python
            video_segment = video_segment.add_keyword("keyword")
            ```
        """
        self.values.append(
            SimpleText.new(
                value=keyword,
                prop_name="hasKeyword",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_keyword_multiple(
        self,
        keywords: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        Add several keywords to the resource

        Args:
            keywords: list of texts
            permissions: optional permissions of these values
            comment: optional comment

        Returns:
            The original resource, with the added keywords

        Examples:
            ```python
            video_segment = video_segment.add_keyword_multiple(["keyword 1", "keyword 2"])
            ```
        """
        vals = check_and_fix_collection_input(keywords, "keywords", self.res_id)
        for v in vals:
            self.add_keyword(v, permissions, comment)
        return self

    def add_keyword_optional(
        self,
        keyword: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        If the value is not empty, add it as keyword, otherwise return the resource unchanged.

        Args:
            keyword: text or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added keyword

        Examples:
            ```python
            video_segment = video_segment.add_keyword_optional("keyword")
            ```

            ```python
            video_segment = video_segment.add_keyword_optional(None)
            ```
        """
        if is_nonempty_value(keyword):
            self.add_keyword(keyword, permissions, comment)
        return self

    def add_relates_to(
        self,
        relates_to: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        Add a link to a related resource

        Args:
            relates_to: ID of the related resource
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added related resource

        Examples:
            ```python
            video_segment = video_segment.add_relates_to("target_resource_id")
            ```
        """
        self.values.append(
            LinkValue.new(
                value=relates_to,
                prop_name="relatesTo",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_relates_to_multiple(
        self,
        relates_to: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        Add several links to related resources

        Args:
            relates_to: list of IDs of the related resources
            permissions: optional permissions of these values
            comment: optional comment

        Returns:
            The original resource, with the added related resources

        Examples:
            ```python
            video_segment = video_segment.add_relates_to_multiple(["target_resource_id_1", "target_resource_id_2"])
            ```
        """
        vals = check_and_fix_collection_input(relates_to, "relatesTo", self.res_id)
        for v in vals:
            self.add_relates_to(v, permissions, comment)
        return self

    def add_relates_to_optional(
        self,
        relates_to: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> VideoSegmentResource:
        """
        If the value is not empty, add it as related resource, otherwise return the resource unchanged.

        Args:
            relates_to: ID or the related resource or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added related resources

        Examples:
            ```python
            video_segment = video_segment.add_relates_to_optional("target_resource_id")
            ```

            ```python
            video_segment = video_segment.add_relates_to_optional(None)
            ```
        """
        if is_nonempty_value(relates_to):
            self.add_relates_to(relates_to, permissions, comment)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> VideoSegmentResource:
        """
        Add metadata from a SALSAH migration.

        Args:
            creation_date: creation date of the resource in SALSAH
            iri: Original IRI in SALSAH
            ark: Original ARK in SALSAH

        Raises:
            InputError: if metadata already exists

        Returns:
            The original resource, with the added metadata

        Examples:
            ```python
            video_segment = video_segment.add_migration_metadata(
                iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
                creation_date="1999-12-31T23:59:59.9999999+01:00"
            )
            ```
        """
        if self.migration_metadata:
            msg_info = MessageInfo(
                "This resource already contains migration metadata, no new data can be added.", resource_id=self.res_id
            )
            raise_input_error(msg_info)
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self


@dataclass
class AudioSegmentResource:
    res_id: str
    label: str
    segment_bounds: SegmentBounds
    values: list[Value]
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        check_and_warn_potentially_empty_string(value=self.label, res_id=self.res_id, expected="string", field="label")

    @staticmethod
    def create_new(
        res_id: str,
        label: str,
        segment_of: str,
        segment_start: float | int | str,
        segment_end: float | int | str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    ) -> AudioSegmentResource:
        """
        Creates a new audio segment resource, i.e. a time span of an AudioRepresentation.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#video-segment-and-audio-segment)

        Args:
            res_id: ID of this audio segment resource
            label: label of this audio segment resource
            segment_of: ID of the audio resource that this segment refers to (cardinality 1)
            segment_start: start of the segment in seconds (cardinality 1)
            segment_end: end of the segment in seconds (cardinality 1)
            permissions: permissions of this resource

        Returns:
            An audio segment resource
        """
        segment_of_val = LinkValue.new(
            value=segment_of, prop_name="isSegmentOf", permissions=permissions, comment=None, resource_id=res_id
        )
        return AudioSegmentResource(
            res_id=res_id,
            label=label,
            segment_bounds=SegmentBounds(segment_start, segment_end, permissions, res_id),
            values=[segment_of_val],
            permissions=permissions,
        )

    def add_title(
        self,
        title: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        Add a title to the resource.

        Args:
            title: text
            permissions: permissions of the value
            comment: comments on the value

        Returns:
            The original resource, with the added title

        Examples:
            ```python
            audio_segment = audio_segment.add_title("segment title")
            ```
        """
        existing_title = [x for x in self.values if x.prop_name == "hasTitle"]
        if any(existing_title):
            _warn_value_exists(
                old_value=existing_title.pop(0).value, new_value=title, value_field="title", res_id=self.res_id
            )
        self.values.append(
            SimpleText.new(
                value=title,
                prop_name="hasTitle",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_title_optional(
        self,
        title: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        If the value is not empty, add it as title, otherwise return the resource unchanged.

        Args:
            title: text or empty value
            permissions: permissions of the value
            comment: comments on the value

        Returns:
            The original resource, with the added title

        Examples:
            ```python
            audio_segment = audio_segment.add_title("segment title")
            ```

            ```python
            audio_segment = audio_segment.add_title(None)
            ```
        """
        if is_nonempty_value(title):
            self.add_title(title, permissions, comment)
        return self

    def add_comment(
        self,
        text: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> AudioSegmentResource:
        """
        Add a comment to the resource

        Args:
            text: text of the comment
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comment

        Examples:
            ```python
            audio_segment = audio_segment.add_comment("comment text")
            ```
        """
        self.values.append(
            Richtext.new(
                value=text,
                prop_name="hasComment",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_comment_multiple(
        self,
        texts: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> AudioSegmentResource:
        """
        Add several comments to the resource

        Args:
            texts: list of texts
            permissions: optional permissions of these values
            comment: optional comment about these comments
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comments

        Examples:
            ```python
            audio_segment = audio_segment.add_comment_multiple(["comment 1", "comment 2"])
            ```
        """
        vals = check_and_fix_collection_input(texts, "hasComment", self.res_id)
        for v in vals:
            self.add_comment(v, permissions, comment, newline_replacement)
        return self

    def add_comment_optional(
        self,
        text: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> AudioSegmentResource:
        """
        If the value is not empty, add it as comment, otherwise return the resource unchanged.

        Args:
            text: text of the comment (or empty value)
            permissions: optional permissions of this value
            comment: optional comment about this comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added comment

        Examples:
            ```python
            audio_segment = audio_segment.add_comment_optional("comment text")
            ```

            ```python
            audio_segment = audio_segment.add_comment_optional(None)
            ```
        """
        if is_nonempty_value(text):
            self.add_comment(text, permissions, comment, newline_replacement)
        return self

    def add_description(
        self,
        description: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> AudioSegmentResource:
        """
        Add a description to the resource

        Args:
            description: text
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added description

        Examples:
            ```python
            audio_segment = audio_segment.add_description("description text")
            ```
        """
        self.values.append(
            Richtext.new(
                value=description,
                prop_name="hasDescription",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_description_multiple(
        self,
        descriptions: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> AudioSegmentResource:
        """
        Add several descriptions to the resource

        Args:
            descriptions: list of texts
            permissions: optional permissions of these values
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added descriptions

        Examples:
            ```python
            audio_segment = audio_segment.add_description_multiple(["description 1", "description 2"])
            ```
        """
        vals = check_and_fix_collection_input(descriptions, "description", self.res_id)
        for v in vals:
            self.add_description(v, permissions, comment, newline_replacement)
        return self

    def add_description_optional(
        self,
        description: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> AudioSegmentResource:
        """
        If the value is not empty, add it as description, otherwise return the resource unchanged.

        Args:
            description: text or empty value
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added description

        Examples:
            ```python
            audio_segment = audio_segment.add_description_optional("description text")
            ```

            ```python
            audio_segment = audio_segment.add_description_optional(None)
            ```
        """
        if is_nonempty_value(description):
            self.add_description(description, permissions, comment, newline_replacement)
        return self

    def add_keyword(
        self,
        keyword: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        Add a keyword to the resource

        Args:
            keyword: text
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added keyword

        Examples:
            ```python
            audio_segment = audio_segment.add_keyword("keyword")
            ```
        """
        self.values.append(
            SimpleText.new(
                value=keyword,
                prop_name="hasKeyword",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_keyword_multiple(
        self,
        keywords: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        Add several keywords to the resource

        Args:
            keywords: list of texts
            permissions: optional permissions of these values
            comment: optional comment

        Returns:
            The original resource, with the added keywords

        Examples:
            ```python
            audio_segment = audio_segment.add_keyword_multiple(["keyword 1", "keyword 2"])
            ```
        """
        vals = check_and_fix_collection_input(keywords, "keywords", self.res_id)
        for v in vals:
            self.add_keyword(v, permissions, comment)
        return self

    def add_keyword_optional(
        self,
        keyword: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        If the value is not empty, add it as keyword, otherwise return the resource unchanged.

        Args:
            keyword: text or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added keyword

        Examples:
            ```python
            audio_segment = audio_segment.add_keyword_optional("keyword")
            ```

            ```python
            audio_segment = audio_segment.add_keyword_optional(None)
            ```
        """
        if is_nonempty_value(keyword):
            self.add_keyword(keyword, permissions, comment)
        return self

    def add_relates_to(
        self,
        relates_to: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        Add a link to a related resource

        Args:
            relates_to: ID of the related resource
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added related resource

        Examples:
            ```python
            audio_segment = audio_segment.add_relates_to("target_resource_id")
            ```
        """
        self.values.append(
            LinkValue.new(
                value=relates_to,
                prop_name="relatesTo",
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_relates_to_multiple(
        self,
        relates_to: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        Add several links to related resources

        Args:
            relates_to: list of IDs of the related resources
            permissions: optional permissions of these values
            comment: optional comment

        Returns:
            The original resource, with the added related resources

        Examples:
            ```python
            audio_segment = audio_segment.add_relates_to_multiple(["target_resource_id_1", "target_resource_id_2"])
            ```
        """
        vals = check_and_fix_collection_input(relates_to, "relatesTo", self.res_id)
        for v in vals:
            self.add_relates_to(v, permissions, comment)
        return self

    def add_relates_to_optional(
        self,
        relates_to: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> AudioSegmentResource:
        """
        If the value is not empty, add it as related resource, otherwise return the resource unchanged.

        Args:
            relates_to: ID of the related resource or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added related resources

        Examples:
            ```python
            audio_segment = audio_segment.add_relates_to_optional("target_resource_id")
            ```

            ```python
            audio_segment = audio_segment.add_relates_to_optional(None)
            ```
        """
        if is_nonempty_value(relates_to):
            self.add_relates_to(relates_to, permissions, comment)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> AudioSegmentResource:
        """
        Add metadata from a SALSAH migration.

        Args:
            creation_date: creation date of the resource in SALSAH
            iri: Original IRI in SALSAH
            ark: Original ARK in SALSAH

        Raises:
            InputError: if metadata already exists

        Returns:
            The original resource, with the added metadata

        Examples:
            ```python
            audio_segment = audio_segment.add_migration_metadata(
                iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
                creation_date="1999-12-31T23:59:59.9999999+01:00"
            )
            ```
        """
        if self.migration_metadata:
            msg_info = MessageInfo(
                "This resource already contains migration metadata, no new data can be added.", resource_id=self.res_id
            )
            raise_input_error(msg_info)
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self


def _check_strings(
    *, string_to_check: str, res_id: str, prop_name: str | None = None, field_name: str | None = None
) -> None:
    if not is_nonempty_value(string_to_check):
        msg_info = MessageInfo("The entered string is not valid.", res_id, prop_name, field_name)
        emit_xmllib_input_warning(msg_info)


def _warn_value_exists(*, old_value: Any, new_value: Any, res_id: str | None, value_field: str | None = None) -> None:
    msg = (
        f"This resource already has a value in this location. "
        f"The old value '{old_value}' is being replace with '{new_value}'."
    )
    msg_info = MessageInfo(msg, res_id, field=value_field)
    emit_xmllib_input_warning(msg_info)



# mypy: disable-error-code="method-assign,no-untyped-def"
import datetime

import pandas as pd
import pytest
import regex

from dsp_tools.error.exceptions import InputError
from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.xmllib.helpers import ListLookup
from dsp_tools.xmllib.helpers import clean_whitespaces_from_string
from dsp_tools.xmllib.helpers import create_footnote_string
from dsp_tools.xmllib.helpers import create_list_from_string
from dsp_tools.xmllib.helpers import create_non_empty_list_from_string
from dsp_tools.xmllib.helpers import create_standoff_link_to_resource
from dsp_tools.xmllib.helpers import create_standoff_link_to_uri
from dsp_tools.xmllib.helpers import escape_reserved_xml_characters
from dsp_tools.xmllib.helpers import find_dates_in_string
from dsp_tools.xmllib.helpers import find_license_in_string
from dsp_tools.xmllib.helpers import make_xsd_compatible_id_with_uuid
from dsp_tools.xmllib.models.config_options import NewlineReplacement
from dsp_tools.xmllib.models.licenses.other import LicenseOther
from dsp_tools.xmllib.models.licenses.recommended import License
from dsp_tools.xmllib.models.licenses.recommended import LicenseRecommended

NBSP = "\u00a0"


@pytest.fixture
def list_lookup() -> ListLookup:
    return ListLookup(
        _lookup={
            "list1": {"Label 1": "list1_node1", "Label 2": "list1_node2"},
            "list2": {"Label 1": "list2_node1", "Label 2": "list2_node2"},
        },
        _prop_to_list_name={
            "default:defaultOntoHasListOne": "list1",
            ":defaultOntoHasListOne": "list1",
            "other-onto:otherOntoHasListOne": "list1",
            "other-onto:otherOntoHasListTwo": "list2",
        },
        _label_language="en",
    )


class TestFootnotes:
    @pytest.mark.parametrize(
        ("input_text", "newline_replacement", "expected"),
        [
            ("Text", NewlineReplacement.NONE, "Text"),
            ("With escape &", NewlineReplacement.NONE, "With escape &amp;"),
            ("With escape <", NewlineReplacement.NONE, "With escape &lt;"),
            ("With escape >", NewlineReplacement.NONE, "With escape &gt;"),
            ("<With escape>", NewlineReplacement.LINEBREAK, "&lt;With escape&gt;"),
            ("Already escaped &lt;&gt;", NewlineReplacement.LINEBREAK, "Already escaped &lt;&gt;"),
            (
                "Non-ASCII àéèêëôû äöüß 0123456789 _-'()+=!?| 漢が글ርبيةб中",
                NewlineReplacement.NONE,
                "Non-ASCII àéèêëôû äöüß 0123456789 _-'()+=!?| 漢が글ርبيةб中",
            ),
        ],
    )
    def test_create_footnote_string_correct(
        self, input_text: str, newline_replacement: NewlineReplacement, expected: str
    ) -> None:
        result = create_footnote_string(input_text, newline_replacement)
        expected = f'<footnote content="{expected}"/>'
        assert result == expected

    @pytest.mark.parametrize(
        ("input_text", "newline_replacement", "expected_msg"),
        [
            (
                "Text",
                NewlineReplacement.PARAGRAPH,
                "Currently the only supported newline replacement is linebreak (<br/>) or None.",
            ),
            (pd.NA, NewlineReplacement.NONE, "The input value is empty."),
        ],
    )
    def test_create_footnote_string_raises(
        self, input_text: str, newline_replacement: NewlineReplacement, expected_msg: str
    ) -> None:
        with pytest.raises(InputError, match=regex.escape(expected_msg)):
            create_footnote_string(input_text, newline_replacement)


def test_create_standoff_link_to_resource_success() -> None:
    result = create_standoff_link_to_resource("id", "Text")
    assert result == '<a class="salsah-link" href="IRI:id:IRI">Text</a>'


def test_create_standoff_link_to_resource_id_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_resource("", "Text")


def test_create_standoff_link_to_resource_text_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_resource("id", "")


def test_create_standoff_link_to_uri_success() -> None:
    result = create_standoff_link_to_uri("https://uri.ch", "Text")
    assert result == '<a href="https://uri.ch">Text</a>'


def test_create_standoff_link_to_uri_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_uri("", "Text")


def test_create_standoff_link_to_uri_text_empty() -> None:
    with pytest.raises(InputError):
        create_standoff_link_to_uri("https://uri.ch", "")


class TestFindDate:
    def test_find_dates_in_string_accept_only_dash_as_range_delimiter(self) -> None:
        assert find_dates_in_string("01.01.1900:31.12.2000") == {
            "GREGORIAN:CE:1900-01-01:CE:1900-01-01",
            "GREGORIAN:CE:2000-12-31:CE:2000-12-31",
        }
        assert find_dates_in_string("01.01.1900 to 31.12.2000") == {
            "GREGORIAN:CE:1900-01-01:CE:1900-01-01",
            "GREGORIAN:CE:2000-12-31:CE:2000-12-31",
        }
        assert find_dates_in_string("1900:2000") == {"GREGORIAN:CE:1900:CE:1900", "GREGORIAN:CE:2000:CE:2000"}
        assert find_dates_in_string("1900 to 2000") == {"GREGORIAN:CE:1900:CE:1900", "GREGORIAN:CE:2000:CE:2000"}

    def test_find_dates_in_string_iso(self) -> None:
        """template: 2021-01-01"""
        assert find_dates_in_string("x 1492-10-12, x") == {"GREGORIAN:CE:1492-10-12:CE:1492-10-12"}
        assert find_dates_in_string("x 0476-09-04. x") == {"GREGORIAN:CE:0476-09-04:CE:0476-09-04"}
        assert find_dates_in_string("x (0476-09-04) x") == {"GREGORIAN:CE:0476-09-04:CE:0476-09-04"}
        assert find_dates_in_string("x [1492-10-32?] x") == set()

    def test_find_dates_in_string_eur_date(self) -> None:
        """template: 31.4.2021 | 5/11/2021 | 2015_01_02"""
        assert find_dates_in_string("x (30.4.2021) x") == {"GREGORIAN:CE:2021-04-30:CE:2021-04-30"}
        assert find_dates_in_string("x (5/11/2021) x") == {"GREGORIAN:CE:2021-11-05:CE:2021-11-05"}
        assert find_dates_in_string("x ...2193_01_26... x") == {"GREGORIAN:CE:2193-01-26:CE:2193-01-26"}
        assert find_dates_in_string("x -2193_01_26- x") == {"GREGORIAN:CE:2193-01-26:CE:2193-01-26"}
        assert find_dates_in_string("x 2193_02_30 x") == set()

    def test_find_dates_in_string_eur_date_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_dates_in_string(f"x 30.4.{cur} x") == {f"GREGORIAN:CE:20{cur}-04-30:CE:20{cur}-04-30"}
        assert find_dates_in_string(f"x 30.4.{nxt} x") == {f"GREGORIAN:CE:19{nxt}-04-30:CE:19{nxt}-04-30"}
        assert find_dates_in_string(f"x 31.4.{nxt} x") == set()

    def test_find_dates_in_string_eur_date_range(self) -> None:
        """template: 27.-28.1.1900"""
        assert find_dates_in_string("x 25.-26.2.0800 x") == {"GREGORIAN:CE:0800-02-25:CE:0800-02-26"}
        assert find_dates_in_string("x 25. - 26.2.0800 x") == {"GREGORIAN:CE:0800-02-25:CE:0800-02-26"}
        assert find_dates_in_string("x 25.-25.2.0800 x") == {"GREGORIAN:CE:0800-02-25:CE:0800-02-25"}
        assert find_dates_in_string("x 25.-24.2.0800 x") == set()

    def test_find_dates_in_string_eur_date_range_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_dates_in_string(f"x 15.-16.4.{cur} x") == {f"GREGORIAN:CE:20{cur}-04-15:CE:20{cur}-04-16"}
        assert find_dates_in_string(f"x 15.-16.4.{nxt} x") == {f"GREGORIAN:CE:19{nxt}-04-15:CE:19{nxt}-04-16"}

    def test_find_dates_in_string_eur_date_range_across_month(self) -> None:
        """template: 26.2.-24.3.1948"""
        assert find_dates_in_string("x _1.3. - 25.4.2022_ x") == {"GREGORIAN:CE:2022-03-01:CE:2022-04-25"}
        assert find_dates_in_string("x (01.03. - 25.04.2022) x") == {"GREGORIAN:CE:2022-03-01:CE:2022-04-25"}
        assert find_dates_in_string("x 28.2.-1.12.1515 x") == {"GREGORIAN:CE:1515-02-28:CE:1515-12-01"}
        assert find_dates_in_string("x 28.2.-28.2.1515 x") == {"GREGORIAN:CE:1515-02-28:CE:1515-02-28"}
        assert find_dates_in_string("x 28.2.-26.2.1515 x") == set()

    def test_find_dates_in_string_eur_date_range_across_month_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_dates_in_string(f"x 15.04.-1.5.{cur} x") == {f"GREGORIAN:CE:20{cur}-04-15:CE:20{cur}-05-01"}
        assert find_dates_in_string(f"x 15.04.-1.5.{nxt} x") == {f"GREGORIAN:CE:19{nxt}-04-15:CE:19{nxt}-05-01"}

    def test_find_dates_in_string_eur_date_range_across_year(self) -> None:
        """template: 1.12.1973 - 6.1.1974"""
        assert find_dates_in_string("x 1.9.2022-3.1.2024 x") == {"GREGORIAN:CE:2022-09-01:CE:2024-01-03"}
        assert find_dates_in_string("x 25.12.2022 - 3.1.2024 x") == {"GREGORIAN:CE:2022-12-25:CE:2024-01-03"}
        assert find_dates_in_string("x 25/12/2022-03/01/2024 x") == {"GREGORIAN:CE:2022-12-25:CE:2024-01-03"}
        assert find_dates_in_string("x 25/12/2022 - 3/1/2024 x") == {"GREGORIAN:CE:2022-12-25:CE:2024-01-03"}
        assert find_dates_in_string("x 25.12.2022-25.12.2022 x") == {"GREGORIAN:CE:2022-12-25:CE:2022-12-25"}
        assert find_dates_in_string("x 25/12/2022-25/12/2022 x") == {"GREGORIAN:CE:2022-12-25:CE:2022-12-25"}
        assert find_dates_in_string("x 25.12.2022-03.01.2022 x") == set()
        assert find_dates_in_string("x 25/12/2022-03/01/2022 x") == set()

    def test_find_dates_in_string_eur_date_range_across_year_2_digit(self) -> None:
        cur = str(datetime.date.today().year - 2000)  # in 2024, this will be "24"
        nxt = str(datetime.date.today().year - 2000 + 1)  # in 2024, this will be "25"
        assert find_dates_in_string(f"x 15.04.23-1.5.{cur} x") == {f"GREGORIAN:CE:2023-04-15:CE:20{cur}-05-01"}
        assert find_dates_in_string(f"x 15.04.{nxt}-1.5.26 x") == {f"GREGORIAN:CE:19{nxt}-04-15:CE:1926-05-01"}

    def test_find_dates_in_string_monthname(self) -> None:
        """template: February 9, 1908 | Dec 5,1908"""
        assert find_dates_in_string("x Jan 26, 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x February26,2051 x") == {"GREGORIAN:CE:2051-02-26:CE:2051-02-26"}
        assert find_dates_in_string("x Sept 1, 1000 x") == {"GREGORIAN:CE:1000-09-01:CE:1000-09-01"}
        assert find_dates_in_string("x October 01, 1000 x") == {"GREGORIAN:CE:1000-10-01:CE:1000-10-01"}
        assert find_dates_in_string("x Nov 6,1000 x") == {"GREGORIAN:CE:1000-11-06:CE:1000-11-06"}

    def test_find_dates_in_string_monthname_after_day(self) -> None:
        """template: 15 Jan 1927 | 15 January 1927"""
        assert find_dates_in_string("x 15 Jan 1927 x") == {"GREGORIAN:CE:1927-01-15:CE:1927-01-15"}
        assert find_dates_in_string("x 15Jan1927 x") == {"GREGORIAN:CE:1927-01-15:CE:1927-01-15"}
        assert find_dates_in_string("x 15 January 1927 x") == {"GREGORIAN:CE:1927-01-15:CE:1927-01-15"}
        assert find_dates_in_string("x 15January1927 x") == {"GREGORIAN:CE:1927-01-15:CE:1927-01-15"}

    def test_find_dates_in_string_german_monthnames(self) -> None:
        """template: 26. Jan. 1993 | 26. Januar 1993"""
        assert find_dates_in_string("x 26 Jan 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26. Jan 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26 Jan. 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26. Jan. 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26.Jan. 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26.Jan.1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26Jan1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}
        assert find_dates_in_string("x 26. Januar 1993 x") == {"GREGORIAN:CE:1993-01-26:CE:1993-01-26"}

    def test_find_dates_in_string_single_year(self) -> None:
        """template: 1907 | 476"""
        assert find_dates_in_string("Text 1848 text") == {"GREGORIAN:CE:1848:CE:1848"}
        assert find_dates_in_string("Text 0476 text") == {"GREGORIAN:CE:476:CE:476"}
        assert find_dates_in_string("Text 476 text") == {"GREGORIAN:CE:476:CE:476"}

    def test_find_dates_in_string_year_range(self) -> None:
        """template: 1849/50 | 1845-50 | 1849/1850"""
        assert find_dates_in_string("x 1849/1850? x") == {"GREGORIAN:CE:1849:CE:1850"}
        assert find_dates_in_string("x 1845-1850, x") == {"GREGORIAN:CE:1845:CE:1850"}
        assert find_dates_in_string("x 800-900, x") == {"GREGORIAN:CE:800:CE:900"}
        assert find_dates_in_string("x 840-50, x") == {"GREGORIAN:CE:840:CE:850"}
        assert find_dates_in_string("x 844-8, x") == {"GREGORIAN:CE:844:CE:848"}
        assert find_dates_in_string("x 1840-1, x") == {"GREGORIAN:CE:1840:CE:1841"}
        assert find_dates_in_string("x 0750-0760 x") == {"GREGORIAN:CE:750:CE:760"}
        assert find_dates_in_string("x 1849/50. x") == {"GREGORIAN:CE:1849:CE:1850"}
        assert find_dates_in_string("x (1845-50) x") == {"GREGORIAN:CE:1845:CE:1850"}
        assert find_dates_in_string("x [1849/1850] x") == {"GREGORIAN:CE:1849:CE:1850"}
        assert find_dates_in_string("x 1850-1849 x") == set()
        assert find_dates_in_string("x 1850-1850 x") == set()
        assert find_dates_in_string("x 830-20 x") == set()
        assert find_dates_in_string("x 830-30 x") == set()
        assert find_dates_in_string("x 1811-10 x") == set()
        assert find_dates_in_string("x 1811-11 x") == set()
        assert find_dates_in_string("x 1811/10 x") == set()
        assert find_dates_in_string("x 1811/11 x") == set()

    @pytest.mark.parametrize("string", ["9 BC", "9 B.C.", "9 BCE", "9 B.C.E."])
    def test_find_dates_in_string_bc_different_notations(self, string: str) -> None:
        assert find_dates_in_string(string) == {"GREGORIAN:BC:9:BC:9"}

    @pytest.mark.parametrize("string", ["9 CE", "9 C.E.", "9 AD", "9 A.D."])
    def test_find_dates_in_string_ce_different_notations(self, string: str) -> None:
        assert find_dates_in_string(string) == {"GREGORIAN:CE:9:CE:9"}

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("x 9 BC x", {"GREGORIAN:BC:9:BC:9"}),
            ("x 10000 BC x", {"GREGORIAN:BC:10000:BC:10000"}),
            ("x 170 BC - 90 BC x", {"GREGORIAN:BC:170:BC:90"}),
            ("x 170-90 BCE x", {"GREGORIAN:BC:170:BC:90"}),
            ("x 20 BCE-50 CE x", {"GREGORIAN:BC:20:CE:50"}),
            ("x 20 BCE - 50 C.E. x", {"GREGORIAN:BC:20:CE:50"}),
        ],
    )
    def test_find_dates_in_string_bc(self, string: str, expected: set[str]) -> None:
        assert find_dates_in_string(string) == expected

    def test_find_dates_in_string_french_bc(self) -> None:
        assert find_dates_in_string("Text 12345 av. J.-C. text") == {"GREGORIAN:BC:12345:BC:12345"}
        assert find_dates_in_string("Text 2000 av. J.-C. text") == {"GREGORIAN:BC:2000:BC:2000"}
        assert find_dates_in_string("Text 250 av. J.-C. text") == {"GREGORIAN:BC:250:BC:250"}
        assert find_dates_in_string("Text 33 av. J.-C. text") == {"GREGORIAN:BC:33:BC:33"}
        assert find_dates_in_string("Text 1 av. J.-C. text") == {"GREGORIAN:BC:1:BC:1"}

    def test_find_dates_in_string_french_bc_ranges(self) -> None:
        assert find_dates_in_string("Text 99999-1000 av. J.-C. text") == {"GREGORIAN:BC:99999:BC:1000"}
        assert find_dates_in_string("Text 1125-1050 av. J.-C. text") == {"GREGORIAN:BC:1125:BC:1050"}
        assert find_dates_in_string("Text 1234-987 av. J.-C. text") == {"GREGORIAN:BC:1234:BC:987"}
        assert find_dates_in_string("Text 350-340 av. J.-C. text") == {"GREGORIAN:BC:350:BC:340"}
        assert find_dates_in_string("Text 842-98 av. J.-C. text") == {"GREGORIAN:BC:842:BC:98"}
        assert find_dates_in_string("Text 45-26 av. J.-C. text") == {"GREGORIAN:BC:45:BC:26"}
        assert find_dates_in_string("Text 53-7 av. J.-C. text") == {"GREGORIAN:BC:53:BC:7"}
        assert find_dates_in_string("Text 6-5 av. J.-C. text") == {"GREGORIAN:BC:6:BC:5"}

    def test_find_dates_in_string_french_bc_orthographical_variants(self) -> None:
        assert find_dates_in_string("Text 1 av. J.-C. text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av J.-C. text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av.J.-C. text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av. J.C. text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av. J-C text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av.JC text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av JC text") == {"GREGORIAN:BC:1:BC:1"}
        assert find_dates_in_string("Text 1 av. J.-C.text") == {"GREGORIAN:BC:1:BC:1"}

    def test_find_dates_in_string_french_bc_dash_variants(self) -> None:
        assert find_dates_in_string("Text 2000-1000 av. J.-C. text") == {"GREGORIAN:BC:2000:BC:1000"}
        assert find_dates_in_string("Text 2000- 1000 av. J.-C. text") == {"GREGORIAN:BC:2000:BC:1000"}
        assert find_dates_in_string("Text 2000 -1000 av. J.-C. text") == {"GREGORIAN:BC:2000:BC:1000"}
        assert find_dates_in_string("Text 2000 - 1000 av. J.-C. text") == {"GREGORIAN:BC:2000:BC:1000"}

    def test_find_dates_in_string_french_bc_invalid_syntax(self) -> None:
        assert find_dates_in_string("Text12 av. J.-C. text") == set()
        assert find_dates_in_string("Text 12 av. J.-Ctext") == set()
        assert find_dates_in_string("Text 1 avJC text") == set()

    @pytest.mark.parametrize(
        "already_parsed",
        [
            "GREGORIAN:BC:2001:BC:2000",
            "GREGORIAN:BC:2001-01:BC:2000-02",
            "GREGORIAN:BC:2001-01-01:BC:2000-01-02",
            "GREGORIAN:BC:1:CE:1",
            "GREGORIAN:CE:1993:CE:1994",
            "GREGORIAN:CE:1993-01:CE:1993-02",
            "GREGORIAN:CE:1993-01-26:CE:1993-01-27",
            "JULIAN:CE:1900:CE:1901",
        ],
    )
    def test_find_dates_in_string_already_parsed(self, already_parsed: str) -> None:
        assert find_dates_in_string(f"text {already_parsed} text") == {already_parsed}

    def test_find_dates_in_string_multiple(self) -> None:
        all_inputs = {
            "01.01.1900-31.12.2000": "GREGORIAN:CE:1900-01-01:CE:2000-12-31",
            "1492-10-12": "GREGORIAN:CE:1492-10-12:CE:1492-10-12",
            "30.4.2021": "GREGORIAN:CE:2021-04-30:CE:2021-04-30",
            "5/11/2021": "GREGORIAN:CE:2021-11-05:CE:2021-11-05",
            "2193_01_26": "GREGORIAN:CE:2193-01-26:CE:2193-01-26",
            "2193_02_30": None,
            "25.-26.2.0800": "GREGORIAN:CE:0800-02-25:CE:0800-02-26",
            "1.3. - 25.4.2022": "GREGORIAN:CE:2022-03-01:CE:2022-04-25",
            "25/12/2022 - 3/1/2024": "GREGORIAN:CE:2022-12-25:CE:2024-01-03",
            "Jan 26, 1993": "GREGORIAN:CE:1993-01-26:CE:1993-01-26",
            "15 January 1927": "GREGORIAN:CE:1927-01-15:CE:1927-01-15",
            "476": "GREGORIAN:CE:476:CE:476",
            "1849/1850": "GREGORIAN:CE:1849:CE:1850",
            "1850/1850": None,
            "1845-1850": "GREGORIAN:CE:1845:CE:1850",
            "844-8": "GREGORIAN:CE:844:CE:848",
            "9 B.C.": "GREGORIAN:BC:9:BC:9",
            "9 AD": "GREGORIAN:CE:9:CE:9",
            "20 BCE - 50 C.E.": "GREGORIAN:BC:20:CE:50",
            "33 av. J.-C.": "GREGORIAN:BC:33:BC:33",
            "842-98 av. J.-C.": "GREGORIAN:BC:842:BC:98",
            "1 av JC": "GREGORIAN:BC:1:BC:1",
            "GREGORIAN:BC:2001:BC:2000": "GREGORIAN:BC:2001:BC:2000",
            "GREGORIAN:BC:2001-01-01:BC:2000-01-02": "GREGORIAN:BC:2001-01-01:BC:2000-01-02",
            "GREGORIAN:CE:1993:CE:1994": "GREGORIAN:CE:1993:CE:1994",
            "GREGORIAN:CE:1993-01-26:CE:1993-01-27": "GREGORIAN:CE:1993-01-26:CE:1993-01-27",
        }
        input_string = " ".join(all_inputs.keys())
        expected = {x for x in all_inputs.values() if x}
        assert find_dates_in_string(input_string) == expected


class TestCreateListFromString:
    def test_create_list_from_string_ok(self) -> None:
        result = create_list_from_string("ab, cd , ", ",")
        assert set(result) == {"ab", "cd"}

    def test_create_list_from_string_not_string(self) -> None:
        msg = regex.escape("The input for this function must be a string. Your input is a bool.")
        with pytest.raises(InputError, match=msg):
            create_list_from_string(True, ",")  # type: ignore[arg-type]

    def test_create_list_from_string_empty(self) -> None:
        result = create_list_from_string(" , ", ",")
        assert isinstance(result, list)
        assert result == []

    def test_create_non_empty_list_from_string_ok(self) -> None:
        result = create_non_empty_list_from_string("ab, cd , ", ",")
        assert set(result) == {"ab", "cd"}

    def test_create_non_empty_list_from_string_raises(self) -> None:
        msg = regex.escape(
            "The input for this function must result in a non-empty list. Your input results in an empty list."
        )
        with pytest.raises(InputError, match=msg):
            create_non_empty_list_from_string(" , ", ",")


class TestListLookup:
    def test_get_node_via_list_name(self, list_lookup):
        assert list_lookup.get_node_via_list_name("list1", "Label 1") == "list1_node1"

    def test_get_node_via_list_name_warns_wrong_list(self, list_lookup):
        msg = regex.escape("The entered list name 'inexistent' was not found. An empty string is returned.")
        with pytest.warns(XmllibInputWarning, match=msg):
            result = list_lookup.get_node_via_list_name("inexistent", "Label 1")
        assert result == ""

    def test_get_node_via_list_name_warns_wrong_node(self, list_lookup):
        msg = regex.escape(
            "'inexistent' was not recognised as label of the list 'list1'. "
            "This ListLookup is configured for 'en' labels."
        )
        with pytest.warns(XmllibInputWarning, match=msg):
            result = list_lookup.get_node_via_list_name("list1", "inexistent")
        assert result == ""

    def test_get_node_via_property(self, list_lookup):
        list_name, node_name = list_lookup.get_list_name_and_node_via_property(
            "other-onto:otherOntoHasListTwo", "Label 2"
        )
        assert list_name == "list2"
        assert node_name == "list2_node2"

    def test_get_node_via_property_warns_wrong_property(self, list_lookup):
        msg = regex.escape("The entered property ':inexistent' was not found. An empty string is returned.")
        with pytest.warns(XmllibInputWarning, match=msg):
            result = list_lookup.get_list_name_and_node_via_property(":inexistent", "Label 2")
        assert result == ("", "")


@pytest.mark.parametrize(
    ("input_val", "expected"),
    [
        ("Text no escape", "Text no escape"),
        ("known tag <strong>content</strong>", "known tag <strong>content</strong>"),
        ("Ampersand &", "Ampersand &amp;"),
        ("Unknow tags <unknown></unknown>", "Unknow tags &lt;unknown&gt;&lt;/unknown&gt;"),
        ("<text in brackets>", "&lt;text in brackets&gt;"),
    ],
)
def test_escape_reserved_xml_characters(input_val: str, expected: str) -> None:
    result = escape_reserved_xml_characters(input_val)
    assert result == expected


@pytest.mark.parametrize(
    ("input_val", "expected"),
    [
        ("Text after space", "Text after space"),
        ("\t Text\nafter newline", "Text after newline"),
        ("\n More text\r    with lots   of spaces    \n", "More text with lots of spaces"),
    ],
)
def test_clean_whitespaces_from_string(input_val: str, expected: str) -> None:
    result = clean_whitespaces_from_string(input_val)
    assert result == expected


def test_clean_whitespaces_from_string_warns() -> None:
    expected = regex.escape("The entered string is empty after all redundant whitespaces were removed.")
    with pytest.warns(XmllibInputWarning, match=expected):
        result = clean_whitespaces_from_string("   \r   \n\t ")
    assert result == ""


class TestFindLicense:
    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text Creative Commons BY 4.0 text", LicenseRecommended.CC.BY),
            ("text Creative Commons BY text", LicenseRecommended.CC.BY),
            ("text CC BY 4.0 text", LicenseRecommended.CC.BY),
            (f"text CC{NBSP}BY{NBSP}4.0 text", LicenseRecommended.CC.BY),
            ("text CC-BY-4.0 text", LicenseRecommended.CC.BY),
            ("text CC_BY_4.0 text", LicenseRecommended.CC.BY),
            ("text CC BY text", LicenseRecommended.CC.BY),
            ("text CC-BY text", LicenseRecommended.CC.BY),
            ("text CC_BY text", LicenseRecommended.CC.BY),
            ("CC BY 4.0", LicenseRecommended.CC.BY),
            ("CC BY", LicenseRecommended.CC.BY),
            ("CC-BY", LicenseRecommended.CC.BY),
        ],
    )
    def test_find_license_different_syntax(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text, CC BY, text", LicenseRecommended.CC.BY),
            (f"text, CC{NBSP}BY, text", LicenseRecommended.CC.BY),
            ("text. CC-BY. text", LicenseRecommended.CC.BY),
            ("text-CC BY-text", LicenseRecommended.CC.BY),
            ("text-CC-BY-text", LicenseRecommended.CC.BY),
            ("text (CC BY) text", LicenseRecommended.CC.BY),
        ],
    )
    def test_find_license_punctuation(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text CC BY text", LicenseRecommended.CC.BY),
            ("text CC BY ND text", LicenseRecommended.CC.BY_ND),
            ("text CC BY SA text", LicenseRecommended.CC.BY_SA),
            ("text CC BY NC text", LicenseRecommended.CC.BY_NC),
            ("text CC BY NC ND text", LicenseRecommended.CC.BY_NC_ND),
            ("text CC BY NC SA text", LicenseRecommended.CC.BY_NC_SA),
            (f"text CC{NBSP}BY{NBSP}NC{NBSP}SA text", LicenseRecommended.CC.BY_NC_SA),
        ],
    )
    def test_find_license_different_licenses(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text cc by text", LicenseRecommended.CC.BY),
            ("text cc by nd text", LicenseRecommended.CC.BY_ND),
            ("text cc by sa text", LicenseRecommended.CC.BY_SA),
            ("text cc by nc text", LicenseRecommended.CC.BY_NC),
            ("text cc by nc nd text", LicenseRecommended.CC.BY_NC_ND),
            ("text cc by nc sa text", LicenseRecommended.CC.BY_NC_SA),
        ],
    )
    def test_find_license_lowercase(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("CC BY NC SA 1.0", LicenseRecommended.CC.BY_NC_SA),
            ("CC BY NC SA 2.0", LicenseRecommended.CC.BY_NC_SA),
            ("CC BY NC SA 3.1", LicenseRecommended.CC.BY_NC_SA),
            ("CC BY NC SA 10.20", LicenseRecommended.CC.BY_NC_SA),
        ],
    )
    def test_find_license_different_versions(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    def test_find_license_ignore_subsequent(self) -> None:
        assert find_license_in_string("CC BY and CC BY SA") == LicenseRecommended.CC.BY

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("CC SA BY 4.0", LicenseRecommended.CC.BY_SA),
            ("CC BY SA NC", LicenseRecommended.CC.BY_NC_SA),
            ("CC BY-ND-NC", LicenseRecommended.CC.BY_NC_ND),
        ],
    )
    def test_find_license_wrong_order(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        "string",
        [
            "CC",
            "CC/BY/SA",
            "CC,BY,SA",
            "CC.BY.SA",
            "textCC-BY-SAtext",
            "CC BY BY",
            "CC BY NC NC",
        ],
    )
    def test_find_license_wrong_format(self, string: str) -> None:
        assert not find_license_in_string(string)

    @pytest.mark.parametrize("string", ["CC ND SA"])
    def test_find_license_non_existent(self, string: str) -> None:
        assert not find_license_in_string(string)

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text AI Generated text", LicenseRecommended.DSP.AI_GENERATED),
            ("text ai generated text", LicenseRecommended.DSP.AI_GENERATED),
            ("text AI text", LicenseRecommended.DSP.AI_GENERATED),
            ("text ai text", LicenseRecommended.DSP.AI_GENERATED),
            ("text IA text", LicenseRecommended.DSP.AI_GENERATED),
            ("text ia text", LicenseRecommended.DSP.AI_GENERATED),
            ("text KI text", LicenseRecommended.DSP.AI_GENERATED),
            ("text ki text", LicenseRecommended.DSP.AI_GENERATED),
            ("textkitext", None),
            ("textKItext", None),
            ("textaitext", None),
            ("textAItext", None),
            ("textiatext", None),
            ("textIAtext", None),
        ],
    )
    def test_find_license_ai_generated(self, string: str, expected: License | None) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text PUBLIC DOMAIN text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text public domain text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            (f"text public{NBSP}domain text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text GEMEINFREI text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text gemeinfrei text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text frei von Urheberrechten text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text urheberrechtsbefreit text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text libre de droits text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("text domaine public text", LicenseRecommended.DSP.PUBLIC_DOMAIN),
        ],
    )
    def test_find_public_domain(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        ("string", "expected"),
        [
            ("text UNKNOWN text", LicenseRecommended.DSP.UNKNOWN),
            ("text unknown text", LicenseRecommended.DSP.UNKNOWN),
            ("text UNBEKANNT text", LicenseRecommended.DSP.UNKNOWN),
            ("text unbekannt text", LicenseRecommended.DSP.UNKNOWN),
            ("text INCONNU text", LicenseRecommended.DSP.UNKNOWN),
            ("text inconnu text", LicenseRecommended.DSP.UNKNOWN),
        ],
    )
    def test_find_unknown(self, string: str, expected: License) -> None:
        assert find_license_in_string(string) == expected

    @pytest.mark.parametrize(
        "string",
        [
            "text Creative Commons 0 1.0 text",
            "text CC 0 1.0 text",
            "text CC-0-1.0 text",
            "text CC-0 text",
            "text CC 0 text",
            "text CC_0_1.0 text",
            f"text CC{NBSP}0 text",
            f"text CC{NBSP}0{NBSP}1.0 text",
        ],
    )
    def test_find_cc_0(self, string: str) -> None:
        assert find_license_in_string(string) == LicenseOther.Public.CC_0_1_0

    @pytest.mark.parametrize(
        "string",
        [
            "text Creative Commons PDM 1.0 text",
            "text CC PDM 1.0 text",
            "text CC-PDM-1.0 text",
            "text CC-PDM text",
            "text CC PDM text",
            "text CC_PDM_1.0 text",
            f"text CC{NBSP}PDM text",
            f"text CC{NBSP}PDM{NBSP}1.0 text",
        ],
    )
    def test_find_public_domain_mark(self, string: str) -> None:
        assert find_license_in_string(string) == LicenseOther.Public.CC_PDM_1_0

    @pytest.mark.parametrize(
        "string",
        [
            "text BORIS Standard License text",
            f"text BORIS{NBSP}Standard{NBSP}License text",
            "text BORIS-Standard-License text",
            "text BORIS_Standard_License text",
            "text Bern Open Repository and Information System Standard License text",
        ],
    )
    def test_find_boris(self, string: str) -> None:
        assert find_license_in_string(string) == LicenseOther.Various.BORIS_STANDARD

    @pytest.mark.parametrize(
        "string",
        [
            "text LICENCE OUVERTE 2.0 text",
            "text licence ouverte text",
            f"text{NBSP}licence{NBSP}ouverte{NBSP}text",
            "text France-licence-ouverte text",
            "text Etalab Licence ouverte text",
        ],
    )
    def test_find_france_ouverte(self, string: str) -> None:
        assert find_license_in_string(string) == LicenseOther.Various.FRANCE_OUVERTE

    @pytest.mark.parametrize(
        ("iri", "lic"),
        [
            ("http://rdfh.ch/licenses/cc-by-4.0", LicenseRecommended.CC.BY),
            ("http://rdfh.ch/licenses/cc-by-nd-4.0", LicenseRecommended.CC.BY_ND),
            ("http://rdfh.ch/licenses/cc-by-nc-4.0", LicenseRecommended.CC.BY_NC),
            ("http://rdfh.ch/licenses/cc-by-nc-nd-4.0", LicenseRecommended.CC.BY_NC_ND),
            ("http://rdfh.ch/licenses/cc-by-nc-sa-4.0", LicenseRecommended.CC.BY_NC_SA),
            ("http://rdfh.ch/licenses/cc-by-sa-4.0", LicenseRecommended.CC.BY_SA),
            ("http://rdfh.ch/licenses/ai-generated", LicenseRecommended.DSP.AI_GENERATED),
            ("http://rdfh.ch/licenses/unknown", LicenseRecommended.DSP.UNKNOWN),
            ("http://rdfh.ch/licenses/public-domain", LicenseRecommended.DSP.PUBLIC_DOMAIN),
            ("http://rdfh.ch/licenses/cc-0-1.0", LicenseOther.Public.CC_0_1_0),
            ("http://rdfh.ch/licenses/cc-pdm-1.0", LicenseOther.Public.CC_PDM_1_0),
            ("http://rdfh.ch/licenses/boris", LicenseOther.Various.BORIS_STANDARD),
            ("http://rdfh.ch/licenses/open-licence-2.0", LicenseOther.Various.FRANCE_OUVERTE),
        ],
    )
    def test_find_license_already_parsed(self, iri: str, lic: License) -> None:
        assert find_license_in_string(iri) == lic

    def test_find_license_edge_cases(self) -> None:
        assert not find_license_in_string("BORIS")  # BORIS alone is not enough; must be followed by "Standard License"


def test_make_xsd_compatible_id() -> None:
    teststring = "0aüZ/_-äöü1234567890?`^':.;+*ç%&/()=±“#Ç[]|{}≠₂₃āṇśṣr̥ṁñἄ𝝺𝝲𝛆’الشعرُאדםПопрыгуньяşğ"  # noqa: RUF001
    _expected = "_0a_Z__-___1234567890_____.__________________________r______________________________"

    result_1 = make_xsd_compatible_id_with_uuid(teststring)
    result_2 = make_xsd_compatible_id_with_uuid(teststring)
    assert result_1 != result_2
    assert result_1.startswith(_expected)
    assert result_2.startswith(_expected)
    assert bool(regex.search(r"^[a-zA-Z_][\w.-]*$", result_1))
    assert bool(regex.search(r"^[a-zA-Z_][\w.-]*$", result_2))

    # test that invalid inputs lead to an error
    with pytest.raises(InputError):
        make_xsd_compatible_id_with_uuid(None)  # type: ignore[arg-type]

    with pytest.raises(InputError):
        make_xsd_compatible_id_with_uuid("")

    with pytest.raises(InputError):
        make_xsd_compatible_id_with_uuid(" ")



from typing import Any
from typing import cast

from loguru import logger

from dsp_tools.commands.project.models.project_definition import ProjectDefinition
from dsp_tools.commands.project.models.project_definition import ProjectMetadata
from dsp_tools.error.exceptions import InputError


def parse_project_json(
    project_json: dict[str, Any],
) -> ProjectDefinition:
    """
    Takes a validated project json and parses the content.

    Args:
        project_json: json of the project

    Returns:
        Object with the parsed information
    """
    metadata = ProjectMetadata(
        shortcode=project_json["project"]["shortcode"],
        shortname=project_json["project"]["shortname"],
        longname=project_json["project"]["longname"],
        keywords=project_json["project"].get("keywords"),
        descriptions=project_json["project"].get("descriptions"),
        enabled_licenses=project_json["project"].get("enabled_licenses"),
    )
    all_lists: list[dict[str, Any]] | None = project_json["project"].get("lists")
    all_ontos = _parse_all_ontos(project_json, all_lists)

    groups = project_json["project"].get("groups")
    users = project_json["project"].get("users")
    return ProjectDefinition(metadata=metadata, ontologies=all_ontos, lists=all_lists, groups=groups, users=users)


def _parse_all_ontos(project_json: dict[str, Any], all_lists: list[dict[str, Any]] | None) -> list[dict[str, Any]]:
    all_ontos: list[dict[str, Any]] = project_json["project"]["ontologies"]
    if all_lists is None:
        return all_ontos
    # rectify the "hlist" of the "gui_attributes" of the properties
    for onto in all_ontos:
        if onto.get("properties"):
            onto["properties"] = _rectify_hlist_of_properties(
                lists=all_lists,
                properties=onto["properties"],
            )
    return all_ontos


def _rectify_hlist_of_properties(
    lists: list[dict[str, Any]],
    properties: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """
    Check the "hlist" of the "gui_attributes" of the properties.
    If they don't refer to an existing list name,
    check if there is a label of a list that corresponds to the "hlist".
    If so, rectify the "hlist" to refer to the name of the list instead of the label.

    Args:
        lists: "lists" section of the JSON project definition
        properties: "properties" section of one of the ontologies of the JSON project definition

    Raises:
        InputError: if the "hlist" refers to no existing list name or label

    Returns:
        the rectified "properties" section
    """

    if not lists or not properties:
        return properties

    existing_list_names = [lst["name"] for lst in lists]

    for prop in properties:
        if not prop.get("gui_attributes"):
            continue
        if not prop["gui_attributes"].get("hlist"):
            continue
        list_name = prop["gui_attributes"]["hlist"] if prop["gui_attributes"]["hlist"] in existing_list_names else None
        if list_name:
            continue

        deduced_list_name = None
        for root_node in lists:
            if prop["gui_attributes"]["hlist"] in root_node["labels"].values():
                deduced_list_name = cast(str, root_node["name"])
        if deduced_list_name:
            msg = (
                f"INFO: Property '{prop['name']}' references the list '{prop['gui_attributes']['hlist']}' "
                f"which is not a valid list name. "
                f"Assuming that you meant '{deduced_list_name}' instead."
            )
            logger.warning(msg)
            print(msg)
        else:
            msg = f"Property '{prop['name']}' references an unknown list: '{prop['gui_attributes']['hlist']}'"
            logger.error(msg)
            raise InputError(f"ERROR: {msg}")
        prop["gui_attributes"]["hlist"] = deduced_list_name

    return properties



from dsp_tools.clients.authentication_client import AuthenticationClient


class AuthenticationClientMockBase(AuthenticationClient):
    server: str
    email: str
    password: str

    def get_token(self) -> str:
        return "mocked_token"



import itertools
import socket
from dataclasses import dataclass

from test.e2e.setup_testcontainers.artifacts import E2E_TESTDATA

TESTCONTAINER_PORTS_LOCKFILES = E2E_TESTDATA / "testcontainer_port_lockfiles"
TESTCONTAINER_PORTS_LOCKFILES.mkdir(parents=True, exist_ok=True)


@dataclass(frozen=True)
class ExternalContainerPorts:
    fuseki: int
    sipi: int
    ingest: int
    api: int


def get_ports() -> ExternalContainerPorts:
    num_of_ports_needed = 4
    port_window: list[int] = []
    for port in itertools.count(1025):
        if len(port_window) == num_of_ports_needed:
            break
        if _reserve_port(port):
            port_window.append(port)

    return ExternalContainerPorts(*port_window)


def _reserve_port(port: int) -> bool:
    with socket.socket() as s:
        if s.connect_ex(("localhost", port)) == 0:
            return False
    try:
        (TESTCONTAINER_PORTS_LOCKFILES / str(port)).touch(exist_ok=False)
    except FileExistsError:
        return False
    return True


def release_ports(ports: ExternalContainerPorts) -> None:
    (TESTCONTAINER_PORTS_LOCKFILES / str(ports.fuseki)).unlink()
    (TESTCONTAINER_PORTS_LOCKFILES / str(ports.sipi)).unlink()
    (TESTCONTAINER_PORTS_LOCKFILES / str(ports.ingest)).unlink()
    (TESTCONTAINER_PORTS_LOCKFILES / str(ports.api)).unlink()



from typing import Any
from typing import Protocol

from dsp_tools.utils.request_utils import PostFiles

# ruff: noqa: D102 (missing docstring in public method)


class Connection(Protocol):
    """
    Protocol class/interface for the connection object.
    """

    def get(
        self,
        route: str,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        pass

    def put(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        pass

    def post(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        files: PostFiles | None = None,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
    ) -> dict[str, Any]:
        pass
